/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/

package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.MappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalReplicationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalSeparationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.IConstrainedEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.Copyable;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.fortiss.af3.task.model.Task;
import org.opt4j.core.Genotype;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * Problem encoding of a task mapping (with redundancy).
 * 
 * @author huang
 * @author diewald
 */
// TODO: The CompositePhenotype is not typed correctly: An additional Type parameter would be needed
// for the subclass of TaskMappingEncoding and passed to the CompositePhenotype.
public abstract class TaskMappingEncoding<S extends ITaskAdapter<?>, T extends TaskMappingEntry<S, IExecutionUnitAdapter<?>>>
		extends MappingEncoding<TaskMappingEncoding<S, T>, S, IExecutionUnitAdapter<?>, T>
		implements IMappingEncoding<S, IExecutionUnitAdapter<?>, T>, Genotype, IExplorationEncoding,
		IConstrainedEncoding, Copyable<TaskMappingEncoding<S, T>> {
	/** Random generator */
	private RandomExploration rand = RandomExploration.getInstance();

	/** Map holding the DSE-internal constraints. */
	protected Multimap<Class<? extends ExplorationConstraint<?>>, ExplorationConstraint<?>> constraintMap =
			HashMultimap.create();

	/** Stores the references between {@link IMappingEntry} to keep track of association updates. */
	protected Multimap<IMappingEntry<?, ?>, IMappingEntry<?, ?>> referencedByMap =
			HashMultimap.create();

	/**
	 * Map of disallowed {@link ITaskAdapter} -> {@link IExecutionUnitAdapter} allocations.
	 */
	// TODO: Move to a central place, but where?
	protected Multimap<ITaskAdapter<?>, IExecutionUnitAdapter<?>> disallowedAllocations;

	/**
	 * Map of disallowed {@link ITaskAdapter} -> {@link IExecutionUnitAdapter} allocations.
	 */
	// TODO: Move to a central place, but where?
	protected Multimap<ITaskAdapter<?>, IExecutionUnitAdapter<?>> allowedAllocations;

	/** Deployable component adapters that shall not be considered by the redundancy operator. */
	protected Collection<ITaskAdapter<?>> fixAllocatedComponents;

	/**
	 * Collection of deployable {@link ITaskAdapter}s represented by {@code this} mapping.
	 */
	protected Collection<S> tasks;

	/** Collection of {@link Partition}s to deploy {@link Task}s to. */
	protected Collection<Partition> partitions;

	/** The input {@link TaskGraphEncoding} that contains the tasks and their signals. */
	protected TaskGraphEncoding<?> tgEncoding;

	/** Ratio between the generation of RoundRobin based and pure random mapping generation. */
	private final double fixToRandomRatio = 0.25;

	/**
	 * Constructor that is initialized with a reference to the DSE specification and the input
	 * system.
	 */
	@SuppressWarnings("unchecked")
	public TaskMappingEncoding(Collection<? extends IExecutionUnitAdapter<?>> resources,
			TaskGraphEncoding<?> tgEncoding) {
		super((Collection<IExecutionUnitAdapter<?>>)resources);
		this.tgEncoding = tgEncoding;
		this.tasks = (Collection<S>)tgEncoding.getTasks();

		disallowedAllocations = createDisallowedAllocationMap();
		allowedAllocations = createAllowedAllocationMap();
		fixAllocatedComponents = createFixAllocationCollection();
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	public TaskMappingEncoding(Collection<? extends IExecutionUnitAdapter<?>> resources,
			TaskGraphEncoding<?> tgEncoding, boolean initialize) throws ExplorationException {
		this(resources, tgEncoding);
		if(initialize) {
			createEncoding(tasks);
		}
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	public TaskMappingEncoding(Collection<IExecutionUnitAdapter<?>> resources,
			TaskGraphEncoding<?> tgEncoding,
			TaskMappingEncoding<ITaskAdapter<?>, ?> incompleteEncoding)
			throws ExplorationException {
		this(resources, tgEncoding);
		completeEncoding(incompleteEncoding);
	}

	/**
	 * Adds the given {@link TaskMappingEntry} to the internal allocation table and adds the
	 * references between {@link TaskMappingEntry}s accordingly.
	 * 
	 * @param entry
	 *            {@link TaskMappingEntry} to be added to the allocation table.
	 */
	protected void addAllocation(T entry) {
		allocationMap.put(entry.getSource(), entry);
		referencedByMap.putAll(entry, entry.getAssociatedElements());
	}

	/**
	 * Creates a collection of component adapters whose redundancy shall not be modified by the
	 * corresponding operator -> define it as fix.
	 */
	// TODO: Exchange this workaround with a mechanism that allows to "reserve" deployable
	// components (-> marker) to not be processed by subsequent operators.
	private Collection<ITaskAdapter<?>> createFixAllocationCollection() {
		Collection<ITaskAdapter<?>> fixAllocatedComponents = new HashSet<>();
		// TODO: DEC_CONSTR
		// for(ITaskAdapter<?> compAdp : tasks) {
		// Optional<ComponentMultiDislocationConstraint> constr =
		// dse.getTargetSpecification().getTargets().stream()
		// .filter(c -> (c instanceof ComponentMultiDislocationConstraint))
		// .map(ComponentMultiDislocationConstraint.class::cast)
		// .filter(c -> c.getComponent() == compAdp.getObject()).findAny();
		// // if(constr.isPresent()) {
		// // fixAllocatedComponents.add(
		// // systemModelAdapter.getDeployableComponentOf(constr.get().getComponent()));
		// // }
		// }
		return fixAllocatedComponents;
	}

	/**
	 * Adds a component adapter whose redundancy shall not be modified by the
	 * corresponding operator -> define it as fix.
	 */
	// TODO: Exchange this workaround with a mechanism that allows to "reserve" deployable
	// components (-> marker) to not be processed by subsequent operators.
	public void addFixAllocation(ITaskAdapter<?> compAdp,
			Collection<IExecutionUnitAdapter<?>> validAllocTargets) {
		fixAllocatedComponents.add(compAdp);
		validAllocTargets.forEach(t -> allowedAllocations.put(compAdp, t));
	}

	/**
	 * Creates a map which refers those {@link IExecutionUnitAdapter}s that are invalid allocation
	 * targets of the given component.
	 */
	private Multimap<ITaskAdapter<?>, IExecutionUnitAdapter<?>> createDisallowedAllocationMap() {
		Multimap<ITaskAdapter<?>, IExecutionUnitAdapter<?>> rval = HashMultimap.create();

		// Collection<ComponentMultiDislocationConstraint> forbiddenDeploymentConstraints =
		// getChildrenWithType(dse, ComponentMultiDislocationConstraint.class);
		//
		// for(ComponentMultiDislocationConstraint currentConstraint :
		// forbiddenDeploymentConstraints) {
		// for(ITaskAdapter<?> currentComponentAdapter : systemModelAdapter
		// .getDeployableComponents()) {
		// if(currentComponentAdapter.getObject() == currentConstraint.getComponent()) {
		// // TODO: Remove dependency on AF3 here.
		// Collection<ExecutionUnit> forbiddenAF3TargetsOfComponent =
		// currentConstraint.getExecutionUnits();
		// for(IExecutionUnitAdapter<?> currentForbiddenTargetUnit : systemModelAdapter
		// .getDeploymentTargets()) {
		// if(forbiddenAF3TargetsOfComponent
		// .contains(currentForbiddenTargetUnit.getObject())) {
		// rval.put(currentComponentAdapter, currentForbiddenTargetUnit);
		// }
		// }
		// }
		// }
		// }

		return rval;
	}

	/**
	 * Creates a lookup map that holds the collection of all disallowed mapping targets of
	 * {@link ITaskAdapter}s.
	 */
	private Multimap<ITaskAdapter<?>, IExecutionUnitAdapter<?>> createAllowedAllocationMap() {
		Multimap<ITaskAdapter<?>, IExecutionUnitAdapter<?>> rval = HashMultimap.create();

		// Collection<ComponentMultiAllocationConstraint> allowedDeploymentConstraints =
		// getChildrenWithType(dse, ComponentMultiAllocationConstraint.class);
		Collection<ITaskAdapter<?>> constrainedComponentAdps = new HashSet<>();

		// for(ComponentMultiAllocationConstraint currentConstraint : allowedDeploymentConstraints)
		// {
		// for(ITaskAdapter<?> currentComponentAdapter : systemModelAdapter
		// .getTasks()) {
		// if(currentComponentAdapter.getObject() == currentConstraint.getComponent()) {
		// constrainedComponentAdps.add(currentComponentAdapter);
		//
		// // TODO: Remove dependency on AF3 here.
		// Collection<ExecutionUnit> allowedAF3TargetsOfComponent =
		// currentConstraint.getExecutionUnits();
		// for(IExecutionUnitAdapter<?> currentAllowedTargetUnit : systemModelAdapter
		// .getDeploymentTargets()) {
		// if(allowedAF3TargetsOfComponent
		// .contains(currentAllowedTargetUnit.getObject())) {
		// rval.put(currentComponentAdapter, currentAllowedTargetUnit);
		// }
		// }
		// }
		// }
		// }

		for(ITaskAdapter<?> curDeplComp : tasks) {
			// TODO: create a field for the deployment targets.
			if(!constrainedComponentAdps.contains(curDeplComp)) {
				getResources().forEach(t -> rval.put(curDeplComp, t));
			}
		}

		return rval;
	}

	/**
	 * Returns the collection of allowed mapping targets if a
	 * {@link ComponentMultiAllocationConstraint} is defined, otherwise {@code null} is returned.
	 */
	// TODO(#2960): Must be based on the DSML expressions.
	public Collection<IExecutionUnitAdapter<?>>
			getAllowedAllocations(ITaskAdapter<?> deployableComponent) {
		// ITaskAdapter<?> origComponent = systemModelAdapter
		// .getDeployableComponentOf((EObject)deployableComponent.getObject());
		// return allowedAllocations.get(origComponent);
		return allowedAllocations.get(deployableComponent);
	}

	/**
	 * Returns the collection of disallowed mapping targets if a
	 * {@link ComponentMultiDislocationConstraint} is defined, otherwise null is returned.
	 */
	// TODO(#2960): Must be based on the DSML expressions.
	public Collection<IExecutionUnitAdapter<?>>
			getDisAllowedAllocations(ITaskAdapter<?> deployableComponent) {
		// ITaskAdapter<?> origComponent = systemModelAdapter
		// .getDeployableComponentOf((EObject)deployableComponent.getObject());
		// return disallowedAllocations.get(origComponent);
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U extends InternalConstraint<?>> Collection<U> getConstraintsOf(Class<U> type) {
		return (Collection<U>)constraintMap.get(type);
	}

	/**
	 * Initializes the allocation table for the {@link ITaskAdapter}s that are not
	 * yet present in the given {@code incompleteEncoding}.
	 * 
	 * @param incompleteEncoding
	 *            Partial Decoding to be completed.
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified.
	 */
	public void completeEncoding(TaskMappingEncoding<?, ?> incompleteEncoding)
			throws ExplorationException {
		ArrayList<S> allocComponents = new ArrayList<>();
		allocComponents.addAll(tasks);

		if(incompleteEncoding != null) {
			allocComponents.removeAll(incompleteEncoding.getRequesters());

			incompleteEncoding.getMappingEntries()
					.forEach(e -> addAllocation(createEntry(e.getSource(), e.getTarget())));
		}

		createEncoding(allocComponents);
	}

	/**
	 * Creates {@link TaskMappingEntry}s for the given set of {@link ITaskAdapter}s.
	 * 
	 * @param allocatableComponents
	 *            Collection of components to generated {@link TaskMappingEntry}s for.
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified.
	 */
	protected void createEncoding(Collection<S> allocatableComponents) throws ExplorationException {
		double selRandOrFixedInit = rand.nextDouble();
		List<IExecutionUnitAdapter<?>> execTargets = new ArrayList<>(getResources());
		int roundRobinAllocIdx = rand.nextInt(execTargets.size());

		List<S> compColl = new ArrayList<>(allocatableComponents);
		if(selRandOrFixedInit < fixToRandomRatio) {
			Collections.shuffle(compColl);
		}

		for(S deployableComponent : compColl) {
			List<Partition> validExecTargets = new ArrayList<>();
			validExecTargets.addAll((Collection<Partition>)(Collection<?>)allowedAllocations
					.get(deployableComponent));
			// TODO: reenable the constraints twds. phys exec Units.
			// if(allowedAllocations.containsKey(deployableComponent)) {
			// List<IExecutionUnitAdapter<?>> physicalExecTargets = allowedAllocations
			// .get(deployableComponent).stream().collect(Collectors.toList());
			// for(IExecutionUnitAdapter<?> physExecUnit : physicalExecTargets) {
			// validExecTargets.addAll(partMappingEncoding.getRequesters(physExecUnit));
			// }
			// }

			if(selRandOrFixedInit < fixToRandomRatio) {
				generateRandomDeployment(deployableComponent);
			} else {
				for(int repl = 0; repl < deployableComponent.getMaxReplication(); repl++) {
					int calcTargetIdx = roundRobinAllocIdx % validExecTargets.size();
					addAllocation(
							createEntry(deployableComponent, validExecTargets.get(calcTargetIdx)));
					roundRobinAllocIdx++;
					if(roundRobinAllocIdx == execTargets.size()) {
						roundRobinAllocIdx = 0;
					}
				}
			}
		}

		// TODO: Readd check with flag.
		// if(!ExplorationAlgDebugUtils.checkTaskMappingForSFConsistency(this)) {
		// System.out.println("Error.");
		// }
	}

	/**
	 * Creates an (encoding specific) mapping entry the describes that a given deployable component
	 * is mapped to a deployment target.
	 */
	protected abstract T createEntry(ITaskAdapter<?> deployableComponent,
			IExecutionUnitAdapter<?> deploymentTarget);

	/**
	 * Creates a {@link TaskMappingEntry} and adds it to {@code this} mapping. The given
	 * {@link TaskMappingEntry} defines the source and the target for the created
	 * {@link TaskMappingEntry}.
	 * 
	 * @param mappingEntry
	 *            Existing {@link TaskMappingEntry} serving as a template for the created
	 *            {@link TaskMappingEntry}.
	 * @return The created {@link TaskMappingEntry}.
	 */
	public <U extends TaskMappingEntry<ITaskAdapter<?>, IExecutionUnitAdapter<?>>> T
			createAndAddEntry(U mappingEntry) {
		T newEntry = createEntry(mappingEntry);
		addAllocation(newEntry);
		return newEntry;
	}

	/**
	 * Creates an (encoding specific) mapping entry that is derived from another
	 * {@link TaskMappingEntry}.
	 * 
	 * @param allocEntry
	 *            {@link TaskMappingEntry} from which the returned {@link TaskMappingEntry} is
	 *            derived.
	 * @return The created {@link TaskMappingEntry}.
	 */
	protected abstract <U extends TaskMappingEntry<ITaskAdapter<?>, IExecutionUnitAdapter<?>>> T
			createEntry(U allocEntry);

	/**
	 * Constructs a random {@TaskMappingEntry} for the given deployable
	 * component. The deployment targets can be given as a list, otherwise the deployment targets
	 * are taken from the platform.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified.
	 */
	public T createRandomEntry(S deployableComponent,
			List<IExecutionUnitAdapter<?>> deploymentTargets) throws ExplorationException {
		IExecutionUnitAdapter<?> target = null;

		// Select from remaining allowed deployment targets if no set of desired deployment targets
		// is given
		if(deploymentTargets == null || deploymentTargets.isEmpty()) {
			deploymentTargets = new ArrayList<IExecutionUnitAdapter<?>>();
			deploymentTargets.addAll(getResources());
			// TODO: Reenable when the constraints twds phys exec units are operational again.
			// for(IResourceAdapter<?> physResource : allowedAllocations.get(deployableComponent)) {
			// deploymentTargets.addAll(partMappingEncoding.getRequesters(physResource));
			// }
		}

		// Remove any disallowed mapping
		deploymentTargets.removeAll(getExcludedDeploymentTargets(deployableComponent));

		if(!deploymentTargets.isEmpty()) {
			target = deploymentTargets.get(rand.nextInt(deploymentTargets.size()));
		}

		if(target == null) {
			throw new ExplorationException(
					"Could not create an allocation entry for the Component " +
							deployableComponent.getName() + ". Typical reasons include:\n" +
							"1. An over-constrained ExplorationSpecification such that there " +
							"exist no valid allocation targets." +
							"2. Desired but non-selected deployment targets in the annoation view " +
							"of the logical platforms" +
							"3. Selection of a Component that shall not be allocated. The selection of " +
							"'deployable' Components is done via the timing model where the 'lowest' timing " +
							"event specifies the layer (for each cluster) of components to allocate.");
		}

		// TODO: This Hack should be removed later on.
		if(target instanceof Partition) {
			((Partition)target).setEmpty(false);
		}

		return createEntry(deployableComponent, target);
	}

	/**
	 * Constructs a random {@TaskMappingEntry} for the given deployable
	 * component. This method respects {@link ComponentMultiDislocationConstraint}s.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified.
	 */
	public T createRandomEntry(S deployableComponent) throws ExplorationException {
		return createRandomEntry(deployableComponent, null);
	}

	/**
	 * Constructs a random {@TaskMappingEntry} for the given deployable
	 * component and adds it to the internal allocation map. This method respects
	 * {@link ComponentMultiDislocationConstraint}s.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified.
	 */
	public T createAndAddRandomEntry(S deployableComponent) throws ExplorationException {
		T entry = createRandomEntry(deployableComponent, null);
		addAllocation(entry);
		return entry;
	}

	/** Removes the given {@code entry} of the given {@code task} from the mapping. */
	@SuppressWarnings({"rawtypes", "unchecked"})
	public void removeEntry(T entry) {
		allocationMap.remove(entry.getSource(), entry);
		referencedByMap.get(entry).forEach(e -> e.rmAssociatedElement((IMappingEntry)entry));
		if(entry.getTarget() instanceof Partition) {
			Partition part = (Partition)entry.getTarget();
			if(getRequesters(part).isEmpty()) {
				part.setEmpty(true);
			}
		}
	}

	/**
	 * Removes a randomly selected entry of the given {@link ITaskAdapter} from the
	 * mapping.
	 */
	public T removeRandomEntry(S deployableComponent) {
		List<T> entries = new ArrayList<>();
		entries.addAll(allocationMap.get(deployableComponent));
		int idx = rand.nextInt(entries.size());
		T rmEntry = entries.get(idx);
		removeEntry(rmEntry);
		return rmEntry;
	}

	/**
	 * Removes the given {@link ITaskAdapter} and, consequently, all associated
	 * entries from the mapping.
	 */
	@SuppressWarnings({"rawtypes", "unchecked"})
	public Collection<T> removeTask(S deployableComponent) {
		for(T entry : allocationMap.get(deployableComponent)) {
			referencedByMap.get(entry).forEach(e -> e.rmAssociatedElement((IMappingEntry)entry));
		}
		return allocationMap.removeAll(deployableComponent);
	}

	// TODO: Do not handle this in the encodings / mutation operators!
	/**
	 * Determines if a deployable component is subject to a
	 * {@link ComponentMultiAllocationConstraint}
	 */
	public boolean hasLimitedDeployment(ITaskAdapter<?> deployableComponent) {
		return allowedAllocations.containsKey(deployableComponent);
	}

	// TODO: Do not handle this in the encodings / mutation operators!
	/**
	 * Generates a fixed deployment for a given deployable component (based on a
	 * {@link ComponentMultiAllocationConstraint} specification)
	 * 
	 * @param deployableComponent
	 *            Deployable component for which fixed deployment is to be generated for
	 */
	public void generateFixedDeployment(S deployableComponent) {
		// Collection<ComponentMultiAllocationConstraint> constraints =
		// dse.getTargetSpecification().getTargets().stream()
		// .filter(t -> (t instanceof ComponentMultiAllocationConstraint))
		// .map(ComponentMultiAllocationConstraint.class::cast)
		// .collect(Collectors.toList());
		// for(ComponentMultiAllocationConstraint cons : constraints) {
		// if(cons.getComponent() == deployableComponent.getObject()) {
		// removeTask(deployableComponent);
		// List<IExecutionUnitAdapter<?>> physExecList = new ArrayList<>(partitions);
		// List<Partition> execList = new ArrayList<>();
		// LambdaUtils.forEach(physExecList, e -> execList.addAll(partMappingEncoding
		// .getRequesters((IResourceAdapter<?>)e)));
		// int allocIdx = 0;
		// for(int i = 0; i < deployableComponent.getMaxReplication(); i++) {
		// allocIdx = rand.nextInt(execList.size());
		// addAllocation(createEntry(deployableComponent, execList.get(allocIdx)));
		// }
		// }
		// }
	}

	/**
	 * Generates a random deployment for a given deployable component (potentially including
	 * redundancy).
	 * 
	 * @param deployableComponent
	 *            Deployable component for which fixed deployment is to be generated for
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified.
	 */
	private void generateRandomDeployment(S deployableComponent) throws ExplorationException {
		// TODO: Model degree of admissible redundancy as a constraint?
		int min = deployableComponent.getMinReplication();
		int max = deployableComponent.getMaxReplication();

		int num = rand.nextInt(max) + min;
		for(int i = 0; i < num; i++) {
			// This respects ExcludeDeploymentConstraints
			// TODO: Generic handling of constraints
			addAllocation(createRandomEntry(deployableComponent));
		}
	}

	/**
	 * Determines the list of deployment target to which a given deployable component may not be
	 * deployed. These restrictions are derived from {@link InternalSeparationConstraint}s and
	 * {@link ComponentMultiDislocationConstraint}s.
	 */
	private Collection<IExecutionUnitAdapter<?>>
			getExcludedDeploymentTargets(ITaskAdapter<?> deployableComponent) {
		Collection<IExecutionUnitAdapter<?>> rval = getDisAllowedAllocations(deployableComponent);

		// Collection<InternalSeparationConstraint> sepConstraints =
		// getConstraintsOf(InternalSeparationConstraint.class);

		// Optional<InternalSeparationConstraint> relevantSepConstraint = sepConstraints.stream()
		// .filter(c -> c.getRequesterSet().contains(deployableComponent)).findAny();

		// TODO: Reinclude for the safety exploration.
		// if(relevantSepConstraint.isPresent()) {
		// // TODO: consider the separation level in the selection of forbidden deployment targets.
		// Collection<IMappingEntry> existingEntries = getIMappingEntriesOf(deployableComponent);
		// existingEntries.forEach(e -> rval.add(e.getTarget()));
		// }

		return rval;
	}

	/** Assigns deployment targets to a deployable component */
	public void setAssignedResources(S deployableComponent, List<T> deploymentTargets) {
		removeTask(deployableComponent);
		deploymentTargets.forEach(e -> addAllocation(e));
	}

	/**
	 * Creates a copy of the internal constraint map of {@code this} {@link TaskMappingEncoding}.
	 */
	// FIXME(AD): This copy method should not be needed anymore. Any task like this should be done
	// generically in the mater
	public Multimap<Class<? extends ExplorationConstraint<?>>, ExplorationConstraint<?>>
			copyConstraintMap(Map<T, T> copyAssocMap) {
		Multimap<Class<? extends ExplorationConstraint<?>>, ExplorationConstraint<?>> copiedMap =
				HashMultimap.create();
		for(Entry<Class<? extends ExplorationConstraint<?>>, ExplorationConstraint<?>> cEntry : constraintMap
				.entries()) {
			ExplorationConstraint<?> cpConstr = null;
			// TODO: generic solution for the copy mechanism.
			if(cEntry.getValue() instanceof InternalSeparationConstraint) {
				cpConstr =
						(ExplorationConstraint<?>)((InternalSeparationConstraint)cEntry.getValue())
								.clone();
			} else if(cEntry.getValue() instanceof InternalReplicationConstraint) {
				cpConstr =
						(ExplorationConstraint<?>)((InternalReplicationConstraint)cEntry.getValue())
								.clone();
			}
			copiedMap.put(cEntry.getKey(), cpConstr);
		}
		return copiedMap;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String str = "";

		for(S deployableComponent : allocationMap.keySet()) {
			str += "Mapping: " + deployableComponent.getName() + " -> " + "{";

			if(allocationMap.get(deployableComponent) != null) {
				for(T tgt : allocationMap.get(deployableComponent)) {
					str += tgt + ", ";
				}
			}
			str += "}\n";
		}
		return str;
	}

	/**
	 * Returns the list of Partitions that are allocated to the given list of physical execution
	 * units. The method is intended to extract a list of partitions that are valid allocation
	 * targets iff the list of physical execution units is defined by the set of mapping constraints
	 * (e.g., derived from available I/Os).
	 */
	// TODO: currently, this is a dummy. Instead constraints should be generated that reference the
	// set of actual targets (here: Partitions) they shall be generated from the input set of
	// constraints and the partition mapping.
	@SuppressWarnings("unchecked")
	public List<Partition> getAllowedPartitionTargets() {
		return (List<Partition>)allocationMap.keySet();
	}

	/**
	 * Returns the size of the encoding
	 */
	@Override
	public int size() {
		return allocationMap.size();
	}
}
