/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene;

import static com.google.common.collect.Maps.newHashMap;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Map;

import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;

import com.google.common.collect.Maps;
import com.google.inject.Provides;

import net.bytebuddy.ByteBuddy;
import net.bytebuddy.description.annotation.AnnotationDescription;
import net.bytebuddy.description.modifier.Visibility;
import net.bytebuddy.dynamic.DynamicType.Builder.FieldDefinition.Optional.Valuable;
import net.bytebuddy.dynamic.DynamicType.Builder.MethodDefinition;
import net.bytebuddy.dynamic.DynamicType.Builder.MethodDefinition.ReceiverTypeDefinition;
import net.bytebuddy.implementation.FieldAccessor;
import net.bytebuddy.implementation.MethodCall;

/**
 * Generates a EncodingProviderModule holding all generated Providers (prototypes). First,
 * the Provider classes get constructed and they are initialized as prototypes such that
 * the @Provides
 * methods are available for binding. At this stage, these providers contain no
 * encoding instances such that provisioning would fail.
 * In order to acquire these instance through the generated providers, the
 * {@link #generateInstanceProviders(Collection)} must be called that
 * creates new Providers holding these instances. For this operation, the prototype module must be
 * available.
 * 
 * @author diewald
 */
public abstract class EncodingProviderGenerator<T> {

	/** Name of the field(s) holding the encoding. */
	protected String fieldName = "encoding";

	/** References the internal prototype provider for concrete encoding providers. */
	protected EncodingProviderModule<T> protoModule;

	/** Constructor. */
	protected EncodingProviderGenerator(Collection<Class<? extends T>> encodingTypesToProvide) {
		this.protoModule = generatePrototypes(encodingTypesToProvide);
	}

	/**
	 * Returns the {@link EncodingProviderModule} that is setup during construction time and
	 * contains the class objects to construct concrete encoding providers.
	 */
	public EncodingProviderModule<T> getEncodingProviderModule() {
		return protoModule;
	}

	/**
	 * Generates a EncodingProviderModule holding all generated prototype Providers. See
	 * {@link EncodingProviderModule} for details.
	 * 
	 * @return EncodingProviderModule holding all generated providers.
	 */
	@SuppressWarnings("unchecked")
	private EncodingProviderModule<T>
			generatePrototypes(Collection<Class<? extends T>> encodingTypesToProvide) {
		Map<Class<T>, EncodingProvider<T>> providersMap = newHashMap();
		for(Class<? extends T> elementType : encodingTypesToProvide) {
			providersMap.put((Class<T>)elementType, createProvider((Class<T>)elementType));
		}
		return new EncodingProviderModule<T>(providersMap);
	}

	/**
	 * Generates a EncodingProviderModule holding all generated Providers. In this step, also the
	 * Providers themselves get created. A pre-constructed prototype {@link EncodingProviderModule}
	 * is required for this operation.
	 * 
	 * @return EncodingProviderModule holding all generated providers.
	 */
	@SuppressWarnings("unchecked")
	public synchronized EncodingProviderModule<T>
			generateInstanceProviders(Collection<T> encodings) {
		Map<Class<T>, EncodingProvider<T>> providersMap = Maps.newHashMap();
		for(T encoding : encodings) {
			EncodingProvider<T> protoProvider = protoModule.getEncodingProviderOf(encoding);
			EncodingProvider<T> instanceProvider = createInstanceOf(
					(Class<? extends EncodingProvider<T>>)protoProvider.getClass(), encoding);
			providersMap.put((Class<T>)encoding.getClass(), instanceProvider);
		}
		return new EncodingProviderModule<T>(providersMap);
	}

	/** Creates a provider for the given encoding. */
	@SuppressWarnings("rawtypes")
	protected EncodingProvider<T> createProvider(Class<T> encodingType) {
		Valuable<EncodingProvider> subClassField = subClassWithField(encodingType);
		ReceiverTypeDefinition<EncodingProvider> withAssigningConstructor =
				createAssigningConstructor(subClassField, encodingType);
		MethodDefinition<EncodingProvider> finishedProvider =
				createProvidesMethod(withAssigningConstructor, encodingType);

		// @CodeFormatterOff
		@SuppressWarnings("unchecked")
		Class<? extends EncodingProvider<T>> providerClass = (Class<? extends EncodingProvider<T>>)
				finishedProvider
				.make()
				.load(getClass().getClassLoader())
				.getLoaded();
		// @CodeFormatterOn

		return createPrototypeOf(providerClass, encodingType);
	}

	/**
	 * Creates a subclass of the {@link EncodingProvider} and adds a field to hold the given
	 * {@code encoding}.
	 */
	@SuppressWarnings("rawtypes")
	protected Valuable<EncodingProvider> subClassWithField(Class<T> encodingType) {
		return new ByteBuddy().subclass(EncodingProvider.class)
				.name("Provides" + encodingType.getSimpleName())
				.defineField(fieldName, encodingType, Visibility.PRIVATE);
	}

	/**
	 * Creates a constructor that assigns the given input encoding to a {@code previous}ly
	 * constructed field.
	 */
	@SuppressWarnings("rawtypes")
	protected ReceiverTypeDefinition<EncodingProvider>
			createAssigningConstructor(Valuable<EncodingProvider> previous, Class<T> encodingType) {
		try {
			// @CodeFormatterOff
			return previous
					.defineConstructor(Visibility.PUBLIC)
					.withParameter(encodingType)
					.intercept(MethodCall
							.invoke(EncodingProvider.class.getDeclaredConstructor())
							.onSuper()
							.andThen(FieldAccessor.ofField(fieldName).setsArgumentAt(0)));
			// @CodeFormatterOn
		} catch(NoSuchMethodException | SecurityException e) {
			throw new ExplorationRuntimeException(
					"Could not generate the setter constructor for the generated" +
							" provider of the encoding " + encodingType.getSimpleName() + ".");
		}
	}

	/**
	 * Creates a getter method for the field named {@link #fieldName} and annotates it with
	 * 
	 * @Provides.
	 */
	@SuppressWarnings("rawtypes")
	protected MethodDefinition<EncodingProvider> createProvidesMethod(
			ReceiverTypeDefinition<EncodingProvider> withConstructor, Class<T> encodingType) {
		// @CodeFormatterOff
		return withConstructor
				.defineMethod("get" + encodingType.getSimpleName(), encodingType,
						Visibility.PUBLIC)
				.intercept(FieldAccessor.ofField(fieldName))
				.annotateMethod(AnnotationDescription.Builder.ofType(Provides.class).build());
		// @CodeFormatterOn
	}

	/**
	 * Creates an instance of the given {@link EncodingProvider} class that wraps the given
	 * {@code encoding}.
	 */
	protected <P extends EncodingProvider<T>> P createInstanceOf(Class<P> providerClass,
			T encoding) {
		try {
			Constructor<P> providerConstructor = providerClass.getConstructor(encoding.getClass());
			// Debug code: To see if the getters are created correctly.
			// for(Method method : providerClass.getDeclaredMethods()) {
			// System.out.println(method.getName());
			// }
			P providerInst = providerConstructor.newInstance(encoding);
			return providerInst;
		} catch(NoSuchMethodException | SecurityException | InstantiationException |
				IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			throw new ExplorationRuntimeException(
					"Could not create an instance of the generated Provider for the Phenotype " +
							encoding.getClass().getSimpleName() + ".",
					e);
		}
	}

	/**
	 * Creates an instance of the given {@link EncodingProvider} class that wraps the given
	 * {@code encoding}.
	 */
	protected EncodingProvider<T> createPrototypeOf(
			Class<? extends EncodingProvider<T>> providerClass, Class<T> encodingType) {
		try {
			Constructor<? extends EncodingProvider<T>> providerConstructor =
					providerClass.getConstructor();
			// Debug code: To see if the getters are created correctly.
			// for(Method method : providerClass.getDeclaredMethods()) {
			// System.out.println(method.getName());
			// }
			EncodingProvider<T> providerInst = providerConstructor.newInstance();
			return providerInst;
		} catch(NoSuchMethodException | SecurityException | InstantiationException |
				IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			throw new ExplorationRuntimeException(
					"Could not create an instance of the generated Provider for the Phenotype " +
							encodingType.getSimpleName() + ".",
					e);
		}
	}
}
