/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph;

import static com.google.common.collect.Sets.newHashSet;

import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

import com.google.common.collect.Sets;
import com.google.inject.Injector;
import com.google.inject.Key;

/**
 * Allows to retrieve the set of required {@link Genotype} that is needed to construct a given set
 * of {@link Phenotype}. Therefore, a visitor is constructed that remembers the bindings to
 * {@link Genotype}s in decode methods. The visitor traverses the decoding {@link Injector} to
 * capture the dependencies of the decoding phase. Here, all potentially used {@link Decoder}s have
 * to be analyzed, i.e. all {@link Decoder}s shipped by the set of enabled
 * {@link IExplorationFeature}s must be bound in the given {@link Injector}.
 * This class is inspired by the InjectorGrapher of the Guice Grapher extension.
 * 
 * @author diewald
 */
public class GenotypeFinder extends BindingWalker<TypeMemorizingDependencyVisitor<Genotype>> {

	/** Constructor. */
	public GenotypeFinder() {
		super(new TypeMemorizingDependencyVisitor<>(newHashSet(Genotype.class)));
	}

	/**
	 * Collects the set of {@link Genotype}s that is required to construct the set of given
	 * {@link Phenotype}s.
	 * 
	 * @param phenotypes
	 *            Set of requested {@link Phenotype}s (e.g., a schedule).
	 * @param injector
	 *            DSE-wide inector to analyze the {@link Decoder} dependencies.
	 * @return Set of required {@link Genotype}s.
	 */
	public Set<Class<? extends Genotype>>
			getGenotypesFor(Set<Class<? extends Phenotype>> phenotypes, Injector injector) {
		Set<Key<?>> phenotypeKeys = Sets.newHashSet();
		phenotypes.forEach(p -> phenotypeKeys.add(Key.get(p)));
		iterateBindings(injector, phenotypeKeys);
		return keyVisitor.getFoundSubTypes();
	}
}
