/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph;

import static com.google.common.collect.Sets.newHashSet;

import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.opt4j.core.problem.Decoder;

import com.google.common.collect.Sets;
import com.google.inject.Injector;
import com.google.inject.Key;

/**
 * Finds the set of required {@link Phenotype}s that are needed to construct a given set
 * of {@link Phenotype}s. Therefore, a visitor is constructed that remembers the bindings to
 * {@link Phenotype}s provided by {@link Decoder}s. The visitor traverses the decoding
 * {@link Injector} to capture the dependencies of the decoding phase. Here, all potentially used
 * {@link Decoder}s have to be analyzed, i.e. all {@link Decoder}s shipped by the set of enabled
 * {@link IExplorationFeature}s must be bound in the given {@link Injector}.
 * This class is inspired by the InjectorGrapher of the Guice Grapher extension.
 * 
 * @author diewald
 */
public class PhenotypeFinder extends BindingWalker<TypeMemorizingDependencyVisitor<Phenotype>> {

	/** Constructor. */
	public PhenotypeFinder() {
		super(new TypeMemorizingDependencyVisitor<>(newHashSet(Phenotype.class)));
	}

	/**
	 * Collects the set of {@link Phenotype}s that are produced by {@link Decoder}s to construct the
	 * set of given {@link Phenotype}s.
	 * 
	 * @param phenotypes
	 *            Set of requested {@link Phenotype}s (e.g., a schedule).
	 * @param injector
	 *            DSE-wide {@link Injector} to analyze the {@link Decoder} dependencies.
	 * @return Set of {@link Phenotype}s constructed during decoding.
	 */
	public Set<Class<? extends Phenotype>>
			getProducedPhenosFor(Set<Class<? extends Phenotype>> phenotypes, Injector injector) {
		Set<Key<?>> phenotypeKeys = Sets.newHashSet();
		phenotypes.forEach(p -> phenotypeKeys.add(Key.get(p)));
		iterateBindings(injector, phenotypeKeys);
		return keyVisitor.getFoundSubTypes();
	}
}
