/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.mating;

import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Queue;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DependentPhenotypeFinder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.PreservablePhenotypeHolder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.service.IExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.module.IRepairModule;
import org.fortiss.tooling.common.util.LambdaUtils;
import org.opt4j.core.Genotype;
import org.opt4j.core.Individual;
import org.opt4j.core.IndividualFactory;
import org.opt4j.core.common.random.Rand;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.operators.copy.Copy;
import org.opt4j.operators.crossover.Crossover;
import org.opt4j.operators.crossover.Pair;
import org.opt4j.operators.mutate.Mutate;
import org.opt4j.operators.mutate.MutationRate;
import org.opt4j.optimizers.ea.Coupler;
import org.opt4j.optimizers.ea.CrossoverRate;
import org.opt4j.optimizers.ea.Mating;
import org.opt4j.optimizers.ea.MatingCrossoverMutate;

import com.google.inject.Inject;

/**
 * {@link Mating} module that extends {@link MatingCrossoverMutate} by providing an additional
 * repair mechanism.
 * 
 * @author diewald
 */
public class MatingCrossoverMutateRepair extends MatingCrossoverMutate {

	/** Holder for {@link Phenotype}s that can be reused in the offspring individual. */
	protected PreservablePhenotypeHolder preservablePhenos;

	/** Finds the {@link Phenotype}s that depend on a modified {@link Genotype}. */
	protected DependentPhenotypeFinder dependPhenoFinder;

	/**
	 * Constructor for the {@link MatingCrossoverMutateRepair} module which extends
	 * {@link MatingCrossoverMutate}.
	 */
	@Inject
	public MatingCrossoverMutateRepair(Crossover<Genotype> crossover, Mutate<Genotype> mutate,
			Copy<Genotype> copy, Coupler coupler, CrossoverRate crossoverRate,
			MutationRate mutationRate, Rand random, IndividualFactory individualFactory,
			PreservablePhenotypeHolder preservablePhenos,
			DependentPhenotypeFinder dependPhenoFinder) {
		super(crossover, mutate, copy, coupler, crossoverRate, mutationRate, random,
				individualFactory);
		this.preservablePhenos = preservablePhenos;
		this.dependPhenoFinder = dependPhenoFinder;
	}

	/** {@inheritDoc} */
	@Override
	protected Pair<Individual> mate(Individual parent1, Individual parent2, boolean doCrossover) {
		Individual repaired1;
		Individual repaired2;
		try {
			repaired1 = repairGenotype(parent1);
			repaired2 = repairGenotype(parent2);
		} catch(ExplorationException e) {
			// Opt4J does not allow to throw Exceptions here.
			throw new RuntimeException(e);
		}

		Pair<Individual> pair;
		if(repaired1 == null && repaired2 == null) {
			pair = super.mate(parent1, parent2, doCrossover);
		} else if(repaired1 == null) {
			pair = super.mate(parent1, repaired2, doCrossover);
		} else if(repaired2 == null) {
			pair = super.mate(repaired1, parent2, doCrossover);
		} else {
			pair = new Pair<Individual>(repaired1, repaired2);
		}
		// DEBUG CODE
		// checkCompositeGenotypeUniqueness(parent1, parent2, pair.getFirst(), pair.getSecond());

		if(parent1.getPhenotype() instanceof Phenotype &&
				parent2.getPhenotype() instanceof Phenotype) {
			updatePreservablePhenos(pair.getFirst().getGenotype(),
					(Phenotype)parent1.getPhenotype());
			updatePreservablePhenos(pair.getSecond().getGenotype(),
					(Phenotype)parent2.getPhenotype());
		}

		return pair;
	}

	/**
	 * Adds those {@link Phenotype}s to the offspring that did not depend on any {@link Genotype}
	 * that was modified such that re-decoding can be omitted.
	 * 
	 * @param genotype
	 *            {@link Genotype} of the offspring {@link Individual}
	 * @param protoPheno
	 *            {@link Phenotype} of the parent offspring {@link Individual}.
	 */
	protected void updatePreservablePhenos(Genotype genotype, Phenotype protoPheno) {
		if(!(protoPheno instanceof ProvidingPhenotype) &&
				!(genotype instanceof CompositeGenotype)) {
			return;
		}

		ProvidingPhenotype pheno = (ProvidingPhenotype)protoPheno;
		CompositeGenotype<?, ?> compGenotype = (CompositeGenotype<?, ?>)genotype;
		// Currently, we modify only one Genotype per iteration. This will change in future.
		Optional<Genotype> modifiedGene = LambdaUtils.getFirst(compGenotype.getModifiedGenotype());
		if(!modifiedGene.isPresent()) {
			throw new ExplorationRuntimeException("No modification was applied to the genotype " +
					genotype.getClass().getSimpleName() + " during this iteration. This " +
					"indicates a programming error, since each offsping genotype shall " +
					"undergo at least one modification to produce a solution alternative " +
					"to existing ones.");
		}

		Collection<Class<? extends Phenotype>> dependentPhenos =
				dependPhenoFinder.getDependentPhenos(modifiedGene.get().getClass());
		Collection<Phenotype> preservablePhenoList = new ArrayList<>();
		for(Entry<Class<? extends Phenotype>, Phenotype> phenoEntry : pheno.getPhenotypeMap()
				.entrySet()) {
			if(!dependentPhenos.contains(phenoEntry.getKey())) {
				preservablePhenoList.add(phenoEntry.getValue());
			}
		}
		preservablePhenos.definePreservablePhenosOf(genotype, preservablePhenoList);
	}

	/**
	 * Triggers the repair method for {@link Individual}s.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	private Individual repairGenotype(Individual ind) throws ExplorationException {
		boolean isRepaired = false;
		Genotype genotype = ind.getGenotype();

		for(IRepairModule<?> repairModule : getService(IExplorationRepairService.class)
				.getRepairFunctionsOf(genotype)) {
			if(repairModule.needsRepair(genotype, ind.getObjectives().getValues())) {
				Genotype copiedGenotype = copy.copy(genotype);
				repairModule.repairGenotype(copiedGenotype, ind.getObjectives().getValues());
				isRepaired = true;
			}
		}

		if(genotype instanceof CompositeGenotype<?, ?>) {
			@SuppressWarnings("unchecked") Queue<Genotype> traversalQueue =
					new LinkedList<>(((CompositeGenotype<?, Genotype>)genotype).values());
			while(!traversalQueue.isEmpty()) {
				Genotype curGenotype = traversalQueue.poll();
				List<IRepairModule<Genotype>> repairModules =
						getService(IExplorationRepairService.class)
								.getRepairFunctionsOf(curGenotype);
				for(IRepairModule<?> curRepairModule : repairModules) {
					if(curRepairModule.needsRepair(curGenotype, ind.getObjectives().getValues())) {
						Genotype copiedGenotype = copy.copy(curGenotype);
						curRepairModule.repairGenotype(copiedGenotype,
								ind.getObjectives().getValues());
						isRepaired = true;
					}
				}

				if(curGenotype instanceof CompositeGenotype) {
					@SuppressWarnings("unchecked") CompositeGenotype<?, Genotype> castedGenotype =
							(CompositeGenotype<?, Genotype>)curGenotype;
					traversalQueue.addAll(castedGenotype.values());
				}
			}
		}

		return isRepaired ? individualFactory.create(genotype) : null;
	}

	/**
	 * Checks whether the genotypes of the offspring refer to differenet objects that the parents.
	 * Debug method.
	 */
	protected void checkCompositeGenotypeUniqueness(Individual p1, Individual p2, Individual o1,
			Individual o2) {
		if(!(p1.getGenotype() instanceof CompositeGenotype)) {
			return;
		}
		for(Object key : ((CompositeGenotype<?, ?>)p1.getGenotype()).keySet()) {
			if(((CompositeGenotype<?, ?>)p1.getGenotype())
					.get(key) == ((CompositeGenotype<?, ?>)o1.getGenotype()).get(key) ||
					((CompositeGenotype<?, ?>)p1.getGenotype())
							.get(key) == ((CompositeGenotype<?, ?>)o2.getGenotype()).get(key) ||
					((CompositeGenotype<?, ?>)p2.getGenotype())
							.get(key) == ((CompositeGenotype<?, ?>)o1.getGenotype()).get(key) ||
					((CompositeGenotype<?, ?>)p2.getGenotype())
							.get(key) == ((CompositeGenotype<?, ?>)o2.getGenotype()).get(key)) {
				throw new RuntimeException(
						"Identified equivalent genotypes in one of the parentsand the offspring");
			}
		}
	}
}
