/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator;

import static org.fortiss.tooling.common.util.LambdaUtils.filterType;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.ExplorationAlgActivator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.Copyable;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.DefaultCopyOperator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.NopCrossover;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.NopMutateOperator;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.tooling.kernel.utils.LoggingUtils;
import org.opt4j.core.Genotype;
import org.opt4j.operators.OperatorModule;
import org.opt4j.operators.copy.CopyModule;
import org.opt4j.operators.crossover.BasicCrossoverModule;
import org.opt4j.operators.crossover.Crossover;
import org.opt4j.operators.crossover.CrossoverModule;
import org.opt4j.operators.mutate.BasicMutateModule;
import org.opt4j.operators.mutate.Mutate;
import org.opt4j.operators.mutate.MutateModule;

/**
 * This module groups mutation operators that affect the mapping of tasks to execution units.
 * 
 * Base class for mutate operator module that enables the registration of pre-constructed operators,
 * which thus can be constructed using a non-default constructor.
 * 
 * @author barner
 */
public class OperatorUtil {

	/**
	 * Returns the {@link OperatorModule} of the given type from the given
	 * {@link IExplorationFeature}. If the given {@link IExplorationFeature} provides multiple
	 * {@link OperatorModule}s, an exception is thrown; if no {@link OperatorModule} of the given
	 * type is provided by the feature, a No-Op {@link OperatorModule} is created.
	 * 
	 * @param operatorType
	 *            type of the {@link OperatorModule}.
	 * @param genotypeType
	 *            modified by the retrieved {@link OperatorModule}.
	 * @param feature
	 *            provides {@link OperatorModule}s.
	 * @return the matching OperatorModule of the given type. Conditions from above apply.
	 * @throws ExplorationException
	 *             if more than one {@link OperatorModule} is defined in the
	 *             {@link IExplorationFeature}.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends OperatorModule<?>, G extends Genotype & Copyable<G>> OperatorModule<?>
			retrieveOperatorModuleType(Class<T> operatorType,
					Class<? extends Genotype> genotypeType, IExplorationFeature feature)
					throws ExplorationException {
		Collection<T> operatorModules =
				filterType(feature.getOperatorsOf(genotypeType), operatorType);
		if(operatorModules.size() > 1) {
			throw new ExplorationException(
					"Only a single mutate module per genotype is currently supported.");
		} else if(operatorModules.size() == 0) {
			// @CodeFormatterOff
			LoggingUtils.warning(
					ExplorationAlgActivator.getDefault(),
					"No " + operatorType.getSimpleName() + " for the genotype " +
					genotypeType.getSimpleName() + " is defined. Installing a NO-OP" +
					" OperatorModule. Expect degraded performance w.r.t. optima search.");
			// @CodeFormatterOn
			// TODO: remove the cast and propagate the Copyable interface to the feature.
			return createNopModule(operatorType, (Class<G>)genotypeType);
		}
		return getFirst(operatorModules).get();
	}

	/**
	 * Creates {@link Crossover} of {@link Mutate} modules performing no modification of the given
	 * {@link Genotype}. Fallback.
	 */
	private static <G extends Genotype & Copyable<G>> OperatorModule<?> createNopModule(
			Class<? extends OperatorModule<?>> operatorType, Class<G> genotypeType) {
		if(MutateModule.class.isAssignableFrom(operatorType)) {
			return new BasicMutateModule() {
				@Override
				protected void configure() {
					super.configure();
					addOperator(genotypeType, NopMutateOperator.class);
				}
			};
		} else if(CrossoverModule.class.isAssignableFrom(operatorType)) {
			return new BasicCrossoverModule() {
				@Override
				protected void configure() {
					super.configure();
					addOperator(genotypeType, NopCrossover.class);
				}
			};
		}
		return null;
	}

	/**
	 * Returns the {@link CopyModule} provided by the given {@link IExplorationFeature} or
	 * constructs a default {@link CopyModule}.
	 * 
	 * @param feature
	 *            provides {@link OperatorModule}s.
	 * @param genotypeType
	 *            copied by the retrieved {@link CopyModule}.
	 * @return CopyModule: Conditions from above apply.
	 * @throws ExplorationException
	 *             if more than one {@link OperatorModule} is defined in the
	 *             {@link IExplorationFeature}.
	 */
	public static <G extends Genotype & Copyable<G>> CopyModule
			retrieveCopyModule(Class<? extends Genotype> genotypeType, IExplorationFeature feature)
					throws ExplorationException {
		Collection<CopyModule> copyModules =
				filterType(feature.getOperatorsOf(genotypeType), CopyModule.class);
		if(copyModules.size() > 1) {
			throw new ExplorationException(
					"Only a single copy operator per genotype is supported.");
		} else if(copyModules.size() == 0) {
			return new CopyModule() {
				@SuppressWarnings("unchecked")
				@Override
				protected void config() {
					addOperator(genotypeType,
							(Class<? extends DefaultCopyOperator<G>>)DefaultCopyOperator.class);
				}
			};
		}
		return getFirst(copyModules).get();
	}
}
