/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution;

import java.util.ArrayList;
import java.util.Collection;
import java.util.TreeMap;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;

/**
 * {@link TreeMap}-based Schedule of a single resource.
 * 
 * @author diewald
 */
public class ResourceSchedule extends TreeMap<TimeSlot, IRequestAdapter<?>> {

	/**
	 * Shifts the given {@link TimeSlot} by the given amount of time and all (temporal) successor
	 * {@link TimeSlot}s of this resource as needed. This means, only those {@link TimeSlot}s are
	 * shifted that would overlap.
	 */
	public Collection<TimeSlot> shiftSlotAndSuccessors(TimeSlot slot, double amount) {
		Collection<TimeSlot> shiftedSlots = new ArrayList<>();
		double prevEndTime = slot.getStartTime() + amount;
		for(TimeSlot succSlot : this.tailMap(slot).keySet()) {
			double currDiff = prevEndTime - succSlot.getStartTime();
			if(currDiff > 0.0) {
				shiftedSlots.add(succSlot);
				succSlot.shift(currDiff);
			} else {
				// If the current slot needs no more shift, successors also don't need a shift.
				// Know by the order by start times.
				break;
			}
			prevEndTime = succSlot.getEndTime();
		}
		return shiftedSlots;
	}

	/**
	 * Indicates whether the temporal order within a {@link ResourceSchedule} is kept. This is
	 * done by evaluating the predecessor field of {@link TimeSlot}s
	 */
	public boolean validateResourceSchedule() {
		for(TimeSlot slot : keySet()) {
			for(TimeSlot pred : slot.getPredecessors()) {
				if(pred.getResource() == slot.getResource() &&
						slot.getStartTime() < pred.getEndTime()) {
					return false;
				}
			}
		}
		return true;
	}
}
