/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static java.lang.Math.pow;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.annotations.Validates;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.constraints.PeriodConstraint;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * This {@link IConstraintEvaluator} evaluates whether all {@link ITaskAdapter}s are
 * executed within their respective period. Any violation is summed and then squared.
 * 
 * @author diewald
 */
public class PeriodConstraintEvaluator<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends EvaluatorModule<Double, PeriodConstraint, StrictTTSchedule<S, T>> {

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public PeriodConstraintEvaluator(Class<PeriodConstraint> optGoal) {
		super(optGoal);
	}

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public PeriodConstraintEvaluator(PeriodConstraint optGoal) {
		super(optGoal);
	}

	/**
	 * evaluate the deadline objective
	 * if the deadline is met, we set the fitness value to -1
	 * if the deadline is violated, we set the fitness value to the gap between the actual latency
	 * and the deadline
	 */
	@Provides
	@Named("PeriodConstraintEvaluator")
	public DoubleValue evaluateGoal(StrictTTSchedule<S, T> schedule,
			InstantiatedTaskGraphEncoding<S> itgEnc, InstantiatedTaskMappingEncoding<?> itmEnc,
			@Named("HyperPeriod") Double hyperPeriod) {
		double retval = 0;

		// we evaluate if all tasks are within the period. we want everything to be
		// finished within the period
		for(ITaskAdapter<?> deployableComponent : itmEnc.getRequesters()) {
			TaskGraph taskGraph = itgEnc.getTaskGraphOf(deployableComponent);
			// we need to check each iteration in the hyper period
			if(taskGraph != null) {
				for(int i = 0; i < hyperPeriod / taskGraph.getPeriod(); i++) {
					double periodEnd = (i + 1) * taskGraph.getPeriod();
					double end = schedule.getLastEndTime(deployableComponent, i);
					if(end > periodEnd) {
						retval += end - periodEnd;
					}
				}
			}
		}

		return new DoubleValue(retval > 0 ? pow(retval, 2) : 0);
	}

	/** Validates the {@link ExplorationConstraint} versus the penalty value. */
	@Validates
	public Boolean validate(Value<?> value) {
		Double evalValue = value.getDouble();
		Double limit = explorationTarget.getLimitation();
		limit = limit == null ? 0 : limit;
		if(evalValue > limit + 10e-6) {
			return false;
		}
		return true;
	}
}
