/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.objective;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.ResourceSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.moea.model.objectives.EnergyMinObjective;
import org.opt4j.core.DoubleValue;

import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Simple energy evaluator that just sums the execution times.
 * 
 * @author huang
 * @author diewald
 */
public class EnergyObjectiveEvaluator<S extends TaskMappingEntry<ITaskAdapter<?>, IExecutionUnitAdapter<?>>, T extends TaskMappingEncoding<ITaskAdapter<?>, S>>
		extends EvaluatorModule<Double, EnergyMinObjective, StrictTTSchedule<S, T>> {

	/** Creates a template energy objective evaluator */
	public EnergyObjectiveEvaluator(Class<EnergyMinObjective> explorationTargetType) {
		super(explorationTargetType);
	}

	/** Creates an energy objective evaluator */
	public EnergyObjectiveEvaluator(EnergyMinObjective explorationTarget) {
		super(explorationTarget);
	}

	/**
	 * This energy evaluator sums the slot lengths of the {@link StrictTTSchedule}. Thus, it
	 * considers all {@link IResourceAdapter}s that are under the regime of that schedule, e.g.,
	 * {@link IExecutionUnitAdapter}s and {@link ITransmissionUnitAdapter}s.
	 */
	@Provides
	@Named("EnergyObjectiveEvaluator")
	public DoubleValue evaluate(StrictTTSchedule<S, T> schedule) {
		if(schedule.getResources().isEmpty()) {
			return new DoubleValue(Double.MAX_VALUE);
		}

		double energy = 0;

		for(IResourceAdapter<?> deploymentTarget : schedule.getResources()) {
			ResourceSchedule localSchedule = schedule.getResourceSchedule(deploymentTarget);
			// Catch the case where no components are deployed on the given resource.
			if(localSchedule != null) {
				for(TimeSlot t : localSchedule.keySet()) {
					assert (t.getDuration() >= 0);
					if(deploymentTarget instanceof IExecutionUnitAdapter) {
						IExecutionUnitAdapter<?> executionUnit =
								(IExecutionUnitAdapter<?>)deploymentTarget;
						ITaskAdapter<?> deployedComponent = t.getDeployedRequester().getSource();

						energy += deployedComponent.getEnergyConsumption(executionUnit);
					} else if(deploymentTarget instanceof ICommunicationResourceAdapter) {
						ICommunicationResourceAdapter<?> commResource =
								(ICommunicationResourceAdapter<?>)deploymentTarget;
						energy += commResource.getPower(t.getDeployedMessage().getMessageSize());
					}
				}
			}
		}

		return new DoubleValue(energy);
	}
}
