/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.af3.exploration.alg.util.AF3Utils.getConnectionTypeDirection;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.isUpdateEnabled;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.setUpdateEnabled;
import static org.fortiss.tooling.common.util.LambdaUtils.filter;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Optional;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IGatewayUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemoryUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.exploration.alg.exception.InvalidPlatformModelException;
import org.fortiss.af3.exploration.dsl_v2.model.expression.SuperSet;
import org.fortiss.af3.platform.hierarchic.model.tile.Tile;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.GatewayUnit;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.MemoryUnit;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.TaskInputPort;
import org.fortiss.af3.task.model.TaskOutputPort;
import org.fortiss.af3.task.model.allocation.ComponentToTaskAllocationEntry;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * {@link SystemModelAdapter} which abstracts AutoFOCUS3 models for the DSE.
 * 
 * @author diewald
 */
public class AF3SystemModelAdapter extends
		SystemModelAdapter<Task, ExecutionUnit, TaskOutputPort, TaskInputPort, GatewayUnit, TransmissionUnit, MemoryUnit> {

	/** References the method which returns the multicast targets (if available). */
	Method multicastGetTargets;

	/** Cache the Annotation View's update status. */
	private boolean annotationUpdateEnabled;

	public static Map<Class<? extends IConnector>, ConnectionType> connectionTypeLUT =
			Collections.unmodifiableMap(new HashMap<Class<? extends IConnector>, ConnectionType>() {
				{
				}
			});

	/** Constructor. */
	public AF3SystemModelAdapter(SuperSet<Task> tasks,
			SuperSet<ComponentToTaskAllocationEntry> componentTaskAllocs,
			SuperSet<ExecutionUnit> execUnits, SuperSet<Signal> signals) throws Exception {
		// FIXME: Porting Code.
		Optional<ComponentToTaskAllocationEntry> anyEntry =
				getFirst(componentTaskAllocs.getEntries());
		if(!anyEntry.isPresent()) {
			throw new Exception("The set of component to task allocations is empty." +
					" It is needed by the DSE. Exiting.");
		}
		Optional<IModelElement> anyComponent = getFirst(anyEntry.get().getSourceElements());
		if(!anyComponent.isPresent()) {
			throw new Exception("At least one task is not associated with a Component." +
					" It is needed by the DSE. Exiting.");
		}
		ComponentArchitecture compArch =
				getFirstParentWithType(anyComponent.get(), ComponentArchitecture.class);
		this.logicalArchitecture = new AF3LogicalArchitectureAdapter(compArch);
		PlatformArchitecture platArch = getFirstParentWithType(
				getFirst(execUnits.getEntries()).orElse(null), PlatformArchitecture.class);
		this.platformArchitecture = new AF3PlatformArchitectureAdapter(platArch);
		// End: Porting code.

		this.taskObjects = tasks.getEntries();
		this.faultContainmentRegion = Tile.class;

		// Disable any annotation view update.
		annotationUpdateEnabled = isUpdateEnabled();
		setUpdateEnabled(false);

		this.deploymentTargets = createDeploymentTargets(execUnits.getEntries());

		memoryUnits = createMemoryUnits(platArch);
		gatewayUnits = createGatewayUnits(platArch);
		transmissionUnits = createTransmissionUnits(platArch);
		availableResources = createAvailableResources();

		platformConnections = createPlatfromResourceConnections(execUnits.getEntries());
		// TODO(#3264): Reintegrate the support for Component Design Libraries.
		// Multimap<Component, ITaskAdapter<Component>> componentToReplacementAdapter =
		// createComponentReplacementAdapters(systemParameters.getReplacementComponents());
		Multimap<Task, ITaskAdapter<Task>> componentToReplacementAdapter = HashMultimap.create();
		this.tasks = createTaskAdapters(componentTaskAllocs, componentToReplacementAdapter);

		// TODO(#3264): The three lines below belong to this issue.
		// componentReplacementAdapters = componentToReplacementAdapter.values();
		nonSafetyFunctionTasks = Collections.emptySet();// new HashSet<>(tasks);
		abstractTasks = Collections.emptySet(); // createAbstractTaskAdapters(tasks);
		this.signals = createSignalAdapters(signals);

		setupTaskGraphs();
		execUnitMapper = createExecUnitMapper();

		setupEmittedMessages();

		// Re-enable annotation view updates.
		setUpdateEnabled(annotationUpdateEnabled);
	}

	/**
	 * Determines the deployable components from an AF3 {@link ComponentArchitecture}.
	 * 
	 * @throws Exception
	 */
	private Collection<ITaskAdapter<Task>> createTaskAdapters(
			SuperSet<ComponentToTaskAllocationEntry> componentTaskAllocs,
			Multimap<Task, ITaskAdapter<Task>> componentReplacementAdapters) throws Exception {
		Collection<ITaskAdapter<Task>> rval = new HashSet<>();
		for(Task currentTask : taskObjects) {
			Optional<ComponentToTaskAllocationEntry> assocComponent =
					getFirst(filter(componentTaskAllocs.getEntries(),
							e -> e.getTargetElement() == currentTask));
			if(!assocComponent.isPresent()) {
				throw new Exception("The task " + currentTask.getName() +
						" is not associated with any Component. This is required for the" +
						" Execution of the DSE.");
			}
			if(assocComponent.get().getSourceElements().size() != 1) {
				throw new Exception("The task " + currentTask.getName() +
						" is associated with none or more than one Components. Currently, it is" +
						" required that each task has exactly one allocated component.");
			}
			Component af3Component =
					(Component)getFirst(assocComponent.get().getSourceElements()).get();
			rval.add(new AF3TaskAdapter(currentTask, af3Component));
		}
		return rval;
	}

	/**
	 * Creates the internal representation of signals exchanged between tasks.
	 * 
	 * @throws InvalidModelException
	 *             if a signal has a sender, but no receiver.
	 */
	private Collection<ISignalAdapter<TaskOutputPort, TaskInputPort>>
			createSignalAdapters(SuperSet<Signal> signals) throws InvalidModelException {
		// TODO: We could also use the signals directly, but that affects the rest of the DSE. Thus
		// this change needs to be considered thoroughly and requires (maybe) large changes.
		Collection<ISignalAdapter<TaskOutputPort, TaskInputPort>> rval = new ArrayList<>();
		Multimap<TaskOutputPort, TaskInputPort> outputPorts = HashMultimap.create();

		for(Signal sig : signals.getEntries()) {
			outputPorts.put((TaskOutputPort)sig.getSource(), (TaskInputPort)sig.getTarget());
		}
		for(TaskOutputPort outPort : outputPorts.keys()) {
			ITaskAdapter<?> senderTask = getTaskAdapter((Task)outPort.getOwner());

			Collection<ITaskAdapter<?>> receiverTasks = new HashSet<>();
			for(TaskInputPort targetIn : outputPorts.get(outPort)) {
				receiverTasks.add(getTaskAdapter((Task)targetIn.getOwner()));
			}
			if(receiverTasks.size() == 0) {
				throw new InvalidModelException((IModelElement)senderTask.getObject(),
						"Found a signal with a sender Task but no receiver. " +
								"This indicates a faulty input Task Architecture.");
			}

			if(!isPureLoopSignal(senderTask, receiverTasks)) {
				// Remove loops from multicast signals if there are any.
				receiverTasks.remove(senderTask);
				rval.add(new AF3SignalAdapter(outPort, senderTask, outputPorts.get(outPort),
						receiverTasks));
			}
		}

		return rval;
	}

	private boolean isPureLoopSignal(ITaskAdapter<?> senderTask,
			Collection<ITaskAdapter<?>> receiverTasks) {
		if(receiverTasks.size() > 1) {
			// A signal with multiple receivers cannot be a "pure" loop: It may send to itself, but
			// this case is handled in the addSignals method.
			return false;
		}

		ITaskAdapter<?> receiver = getFirst(receiverTasks).get();
		if(senderTask == receiver) {
			return true;
		}
		return false;
	}

	/**
	 * Creates {@link IMemoryUnitAdapter}s from AF3-{@link MemoryUnit}s.
	 * 
	 * @throws Exception
	 */
	private Collection<IMemoryUnitAdapter<MemoryUnit>>
			createMemoryUnits(PlatformArchitecture platformArchitecture) throws Exception {
		Collection<IMemoryUnitAdapter<MemoryUnit>> rval =
				new ArrayList<IMemoryUnitAdapter<MemoryUnit>>();
		for(MemoryUnit currentMemoryUnit : getChildrenWithType(platformArchitecture,
				MemoryUnit.class)) {
			rval.add(new AF3MemoryUnitAdapter(currentMemoryUnit));
		}
		return rval;
	}

	/**
	 * Determines the deployment targets from an AF3 {@link PlatformArchitecture}.
	 * 
	 * @throws Exception
	 */
	// TODO(#3262): Reintegrate failure rates.
	private Collection<IExecutionUnitAdapter<ExecutionUnit>> createDeploymentTargets(
			Collection<ExecutionUnit> targetExecutionUnits) throws Exception {
		if(targetExecutionUnits.isEmpty()) {
			throw new Exception(
					"No deployment targets have been specified in the target platform architecture(s). Please check whether the correct platform architecture was selected and whether the desired Execution Units were selected as deployment targets (Platform Model --> Annotation View --> Deployment Target)");
		}
		Collection<IExecutionUnitAdapter<ExecutionUnit>> rval =
				new ArrayList<IExecutionUnitAdapter<ExecutionUnit>>();
		for(ExecutionUnit execUnit : targetExecutionUnits) {
			rval.add(new AF3ExecutionUnitAdapter(execUnit));
		}
		return rval;
	}

	/**
	 * Collects all {@link GatewayUnit}s from a given AF3 Platform and returns them as a collection
	 * of {@link IGatewayUnitAdapter}s.
	 * 
	 * @throws Exception
	 */
	// TODO(#3262): Also power here...
	private Collection<IGatewayUnitAdapter<GatewayUnit>>
			createGatewayUnits(PlatformArchitecture platformArchitecture) throws Exception {
		Collection<IGatewayUnitAdapter<GatewayUnit>> rval =
				new ArrayList<IGatewayUnitAdapter<GatewayUnit>>();

		for(GatewayUnit gatewayUnit : EcoreUtils.getChildrenWithType(platformArchitecture,
				GatewayUnit.class)) {
			rval.add(new AF3GatewayAdapter(gatewayUnit));
		}

		return rval;
	}

	/**
	 * Determines the communication resource AF3 {@link PlatformArchitecture}.
	 * 
	 * @throws Exception
	 */
	// TODO(#3262): Also power, here.
	private Collection<ITransmissionUnitAdapter<TransmissionUnit>>
			createTransmissionUnits(PlatformArchitecture platformArchitecture) throws Exception {
		Collection<ITransmissionUnitAdapter<TransmissionUnit>> rval =
				new ArrayList<ITransmissionUnitAdapter<TransmissionUnit>>();

		for(TransmissionUnit transmissionUnit : EcoreUtils.getChildrenWithType(platformArchitecture,
				TransmissionUnit.class)) {
			rval.add(new AF3TransmissionUnitAdapter(transmissionUnit));
		}
		return rval;
	}

	// TODO: Consider removing "available resources", since this might be already implemented in the
	// virtual HW mapping
	// TODO: Rebase createAvailableResources() upon the IPlatformResource, since this the
	// appropriate marker interface. Only create new IResourceAdapters, if no other Resource is
	// present already.
	/**
	 * Creates a {@link Collection} that references all resources present in the given platform.
	 * Currently, those are {@link ExecutionUnit}s, {@link GatewayUnit}s, and
	 * {@link TransmissionUnit}s.
	 */
	private Collection<IResourceAdapter<?>> createAvailableResources() {
		Collection<IResourceAdapter<?>> rval = new ArrayList<IResourceAdapter<?>>();
		rval.addAll(deploymentTargets);
		rval.addAll(memoryUnits);
		rval.addAll(gatewayUnits);
		rval.addAll(transmissionUnits);
		return rval;
	}

	/**
	 * Creates the {@link IResourceConnectionAdapter}s that describe the connections present in the
	 * {@link PlatformArchitecture}.
	 * 
	 * @throws Exception
	 *             if the end point of a connection cannot be determined.
	 */
	private Collection<IResourceConnectionAdapter> createPlatfromResourceConnections(
			Collection<ExecutionUnit> af3DeploymentTargets) throws Exception {
		// Initial traversal list of IConnectors attached to ExecutionUnits selected as deployment
		// targets.
		LinkedList<IConnector> deployableConnectors = new LinkedList<IConnector>();
		for(IExecutionUnitAdapter<?> execUnit : deploymentTargets) {
			deployableConnectors.addAll(((ExecutionUnit)execUnit.getObject()).getConnectors());
		}

		// TODO: rewrite the code below. Handling of hull elements and virtual targets should be
		// considered in a single util method. Based on the collected target connectors, the
		// resourceconnectionadapters should be generated.
		Collection<IConnector> visitedConnectors = new HashSet<IConnector>();
		Collection<IResourceConnectionAdapter> rval = new HashSet<IResourceConnectionAdapter>();
		while(!deployableConnectors.isEmpty()) {
			IConnector currentConnector = deployableConnectors.poll();
			visitedConnectors.add(currentConnector);
			// NOTE: Here, source element does not imply "source of the connection", but
			// "source of the traversal".
			IHierarchicElement sourceElement = currentConnector.getOwner();
			if(sourceElement instanceof ExecutionUnit || sourceElement instanceof GatewayUnit ||
					sourceElement instanceof TransmissionUnit) {
				// Path walk to determine the *real* connection targets, i.e. ignore hull elements
				// like Tiles whose contained Cores (or partitions) were selected as deployment
				// targets.
				Collection<IConnector> targetConnectors =
						getEffectiveConnectionTargets(currentConnector, af3DeploymentTargets);

				// Create an IResourceConnectionAdapter between the sourceElement of the traversal
				// and the *real* targets.
				AF3ResourceConnectionAdapter connAdp =
						createResourceConnection(currentConnector, targetConnectors);
				if(connAdp != null) {
					rval.add(connAdp);
				}
				visitedConnectors.addAll(targetConnectors);

				// Collect the IConnectors attached to the target elements of the created
				// IResourceConnectionAdapter for further traversing the PlatformArchitecture.
				//
				// TODO Connections between IModelElements of different PlatformArchitectures which
				// used to be realized via ResourceLink specifications were automatically considered
				// (ResourceLink has been removed in favor of af3.allocation).
				Collection<IHierarchicElement> targetElements = new HashSet<IHierarchicElement>();
				for(IConnector currentTargetConnector : targetConnectors) {
					targetElements.add(currentTargetConnector.getOwner());
				}
				determineNextConnectionTargets(deployableConnectors, visitedConnectors,
						targetElements, af3DeploymentTargets);
			}
		}

		return rval;
	}

	/**
	 * Collect the IConnectors attached to the target elements of the created
	 * IResourceConnectionAdapter for further traversing the PlatformArchitecture.
	 */
	// TODO: Refactor this method to use the static methods from the AF3Utils class instead.
	private void determineNextConnectionTargets(LinkedList<IConnector> deployableConnectors,
			Collection<IConnector> visitedConnectors, Collection<IHierarchicElement> targetElements,
			Collection<ExecutionUnit> af3DeploymentTargets) {
		for(IHierarchicElement targetElement : targetElements) {
			if(targetElement instanceof IPlatformResource &&
					!af3DeploymentTargets.contains(targetElement)) {
				for(IConnector connectorAtTarget : targetElement.getConnectors()) {
					// Do not traverse back again.
					if(connectorAtTarget != targetElement) {
						if(!visitedConnectors.contains(connectorAtTarget)) {
							deployableConnectors.add(connectorAtTarget);
						}
					}
				}
			}
		}
	}

	/**
	 * Creates a {@link IResourceConnectionAdapter} between the given source and the target
	 * elements. This method takes {@link IHierarchicElement}s as inputs and identifies the matching
	 * {@link IResourceConnectionAdapter}s of the given elements.
	 * 
	 * @throws InvalidPlatformModelException
	 *             if an invalid combination of of platform ports/interfaces has been detected
	 *             (e.g., Input <-> Input).
	 */
	private AF3ResourceConnectionAdapter createResourceConnection(IConnector sourceConnector,
			Collection<IConnector> targetConnectors) throws InvalidPlatformModelException {
		IResourceAdapter<?> sourceResource = null;
		Collection<IResourceAdapter<?>> targetResources = new HashSet<IResourceAdapter<?>>();
		Map<IResourceAdapter<?>, ConnectionType> targetConnType =
				new HashMap<IResourceAdapter<?>, ConnectionType>();
		Collection<IHierarchicElement> targetElements = new HashSet<IHierarchicElement>();

		// Capture information about the target elements and the IConnectors that are the actual AF3
		// targets of this connection. It is used to determine the connection direction for each
		// target.
		Multimap<IHierarchicElement, IConnector> targetConnsAtElement = HashMultimap.create();
		for(IConnector targetConnector : targetConnectors) {
			targetElements.add(targetConnector.getOwner());
			targetConnsAtElement.put(targetConnector.getOwner(), targetConnector);
		}

		// Identify the corresponding IResourceAdapters of the collection of target
		// IHierarchicalElements and determine the direction of their connection.
		for(IResourceAdapter<?> resource : availableResources) {
			if(resource.getObject() == sourceConnector.getOwner()) {
				sourceResource = resource;
			} else if(targetElements.contains(resource.getObject())) {
				targetResources.add(resource);
				ConnectionType type = ConnectionType.BIDIRECTIONAL;
				boolean firstRoundFlag = true;
				for(IConnector targetConnector : targetConnsAtElement
						.get((IHierarchicElement)resource.getObject())) {
					if(firstRoundFlag) {
						firstRoundFlag = false;
						type = getConnectionTypeDirection(sourceConnector, targetConnector, this);
					} else if(type != getConnectionTypeDirection(sourceConnector, targetConnector,
							this)) {
						type = ConnectionType.BIDIRECTIONAL;
					}
				}
				targetConnType.put(resource, type);
			}
		}
		assert (sourceResource != null);

		return new AF3ResourceConnectionAdapter(sourceResource, targetConnType);
	}

	/**
	 * Returns the (first) {@link ITaskAdapter} which is associated with the given
	 * AF3 {@link Component}. Returns null if no matching {@link ITaskAdapter} is
	 * found.
	 */
	public ITaskAdapter<?> getTaskAdapter(Task af3Task) {
		if(af3Task == null) {
			throw new IllegalArgumentException("Cannot identify ITaskAdapter for 'null' Tasks");
		}
		ITaskAdapter<?> taskAdapter = null;
		for(ITaskAdapter<Task> task : tasks) {
			if(task.getObject() == af3Task) {
				taskAdapter = task;
				break;
			}
		}
		return taskAdapter;
	}

	/**
	 * Returns the (first) {@link IExecutionUnitAdapter} which is associated with the given
	 * AF3 {@link ExecutionUnit}. Returns null if no matching {@link IExecutionUnitAdapter} is
	 * found
	 */
	public IExecutionUnitAdapter<?> getExecutionUnitAdapter(ExecutionUnit af3ExecutionUnit) {
		if(af3ExecutionUnit == null)
			throw new IllegalArgumentException(
					"Cannot identify IExecutionUnitAdapter for 'null' ExecutionUnits");

		IExecutionUnitAdapter<?> executionUnitAdapter = null;
		for(IExecutionUnitAdapter<?> targetExecutionUnit : deploymentTargets) {
			if(targetExecutionUnit.getObject() == af3ExecutionUnit) {
				executionUnitAdapter = targetExecutionUnit;
				break;
			}
		}

		assert (executionUnitAdapter == null) : "No IExecutionUnitAdapter found for the ExecutionUnit " +
				af3ExecutionUnit.getName() + "(ID: " + af3ExecutionUnit.getId() +
				"). Every AF3-ExecutionUnit should have a IExecutionUnitAdapter equivalent.";

		return executionUnitAdapter;
	}

	/** Determine task graph for a given component */
	public TaskGraph getTaskGraph(Task task) {
		// TODO: AD: The cast to the AF3 systemmodeladapter breaks the abstration the goals. It
		// would be rather required to define the source and sink components as
		// IDeployableComponents in the ecore model. The transformation from AF3 components to the
		// IDeployableCompontes should be done when the data from the GUI is written to the model.
		ITaskAdapter<?> taskAdapter = getTaskAdapter(task);
		return getTaskGraph(taskAdapter);
	}
}
