/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.af3.exploration.alg.util.AF3Utils.getAnnotationChecked;

import java.math.BigDecimal;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;

import org.eclipse.emf.common.util.Enumerator;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.exploration.moea.model.annotation.ReplicationBounds;
import org.fortiss.af3.exploration.projectmodel.DSE;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.af3.project.utils.ProjectUtils;
import org.fortiss.af3.safety.model.annotation.SafetyIntegrityLevel;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.timing.TaskStartEvent;
import org.fortiss.af3.timing.model.TimingSpecification;
import org.fortiss.af3.timing.utils.TimingUtils;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Adapter between the AF3 deployable {@link Component} (=tasks) and the DSE.
 * 
 * @author barner
 */
public class AF3TaskAdapter extends AF3RequestAdapter<Task> implements ITaskAdapter<Task> {
	/** References the realized AF3 component. */
	// FIXME: Remove after updating the transformation framework for SuperSets.
	private Component af3Component;

	/** Map of possible Wcets defined by the deployment target. */
	@SuppressWarnings("unused")
	private Map<IExecutionUnitAdapter<?>, Double> wcetMap;

	/** Map of possible energy consumption defined by the deployment target. */
	private Map<IExecutionUnitAdapter<?>, Double> energyMap;

	/** Minimum number of replica (1 == no replication, just the component itself). */
	private int minReplication;

	/** Maximum number of replica (1 == no replication, just the component itself). */
	private int maxReplication;

	/** Period annotation */
	private Double period;

	/** SIL annotation */
	private SafetyIntegrityLevel sil;

	/** Reference to the corresponding elements of the diverse implementations. */
	private Collection<ITaskAdapter<Component>> replacementComponents;

	/**
	 * "Trace" to the {@link ITaskAdapter} which {@code this} {@link ITaskAdapter} replaced (if
	 * applicable).
	 */
	private ITaskAdapter<Component> replacedComponent;

	/**
	 * Creates an {@link ITaskAdapter} for a {@link Task}.
	 * 
	 * @throws InvalidModelException
	 *             if a required annotation cannot be found.
	 */
	public AF3TaskAdapter(Task af3Task, Component af3Component) throws InvalidModelException {
		super(af3Task);

		// BigDecimal af3Period = TimingUtils.getPeriod(ts, element, eventType);
		FileProject fp = ProjectUtils.getFileProject(af3Task);
		DSE dseNode = EcoreUtils.getFirstChildWithType(fp, DSE.class);
		TimingSpecification ts = dseNode.getTimingSpecification();
		BigDecimal af3Period = TimingUtils.getPeriod(ts, af3Task, TaskStartEvent.class);
		if(af3Period == null || af3Period.doubleValue() <= 0.0) {
			throw new InvalidModelException(af3Task,
					"A non-specified period value and negative periods are not allowed.");
		}
		period = af3Period.doubleValue();
		sil = getAnnotationChecked(af3Component, SafetyIntegrityLevel.class);
		ReplicationBounds replicationSpec = getAnnotationChecked(af3Task, ReplicationBounds.class);
		minReplication = replicationSpec.getMinReplication();
		maxReplication = replicationSpec.getMaxReplication();

		// FIXME: Remove after updating the transformation framework for SuperSets.
		this.af3Component = af3Component;

		replacementComponents = new HashSet<ITaskAdapter<Component>>();
	}

	// TODO(#3264): Remove this method, component adapters are used.
	public AF3TaskAdapter(Task task, Component af3Component,
			Collection<ITaskAdapter<Component>> componentReplacementAdapters)
			throws InvalidModelException {
		this(task, af3Component);
		replacementComponents = componentReplacementAdapters;
	}

	/**
	 * Creates an "concrete" {@link ITaskAdapter} for a {@link Task} with the given period.
	 * 
	 * @throws InvalidModelException
	 *             if a required annotation cannot be found.
	 */
	public AF3TaskAdapter(Task task, Component af3Component, ITaskAdapter<Task> replacedComponent)
			throws InvalidModelException {
		this(task, af3Component);
		// this.replacedComponent = replacedComponent;
	}

	/** {@inheritDoc} */
	@Override
	public Double getWcet(IExecutionUnitAdapter<?> executionUnit) {
		// TODO (#3262): HACK More flexible approach is needed: We distinguish between phys. exec
		// Units and Partitions.
		return 10.0;
		// return wcetMap.get(executionUnit);
	}

	/** {@inheritDoc} */
	@Override
	public Double getEnergyConsumption(IExecutionUnitAdapter<?> executionUnit) {
		return energyMap.get(executionUnit);
	}

	/** {@inheritDoc} */
	@Override
	public Enumerator getRequiredSafetyIntegrityLevel() {
		return sil.getValue();
	}

	/** {@inheritDoc} */
	@Override
	public Double getPeriod() {
		return period;
	}

	/** {@inheritDoc} */
	@Override
	public void setPeriod(double period) {
		this.period = period;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isInstantiationRequired() {
		return !replacementComponents.isEmpty();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isInstantiation() {
		return replacedComponent != null;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<ITaskAdapter<Task>> getReplacementComponents() {
		// TODO(#3264)
		// return replacementComponents;
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public ITaskAdapter<Task> getReplacedComponent() {
		// TODO(#3264)
		// return replacedComponent;
		return null;
	}

	/** {@inheritDoc} */
	// TODO: If possible, find a way to avoid these suppresswarnings.
	@SuppressWarnings("unchecked")
	@Override
	public void setReplacedComponent(@SuppressWarnings("rawtypes") ITaskAdapter abstractComponent) {
		replacedComponent = abstractComponent;
	}

	/** {@inheritDoc} */
	@Override
	public void updateInstatiationData() {
		replacedComponent = null;
	}

	/** {@inheritDoc} */
	@Override
	public int getMinReplication() {
		return minReplication;
	}

	/** {@inheritDoc} */
	@Override
	public int getMaxReplication() {
		return maxReplication;
	}

	/** {@inheritDoc} */
	@Override
	public AF3TaskAdapter clone() {
		try {
			return (AF3TaskAdapter)super.clone();
		} catch(CloneNotSupportedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Component getAssignedComponent() {
		return af3Component;
	}
}
