/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/

package org.fortiss.af3.exploration.alg.guava;

import static org.conqat.lib.commons.reflect.ReflectionUtils.performNearestClassLookup;

import java.util.HashMap;
import java.util.Map;

import com.google.common.collect.ForwardingMap;
import com.google.common.primitives.Primitives;

/**
 * 
 * Modifiable version of this {@link SubClassToInstanceMap}.
 * 
 * @author diewald
 */
// using writeReplace instead of standard serialization
public final class MutableSubClassToInstanceMap<B> extends ForwardingMap<Class<? extends B>, B>
		implements SubClassToInstanceMap<B> {

	/** Backend map to hold the actual content. */
	protected Map<Class<? extends B>, B> delegate;

	/**
	 * Returns a new {@code MutableClassToInstanceMap} instance backed by a {@link HashMap} using
	 * the default initial capacity and load factor.
	 */
	public static <B> MutableSubClassToInstanceMap<B> create() {
		return new MutableSubClassToInstanceMap<B>(new HashMap<Class<? extends B>, B>());
	}

	/**
	 * Returns a new {@code MutableClassToInstanceMap} instance backed by a given
	 * empty {@code backingMap}. The caller surrenders control of the backing map,
	 * and thus should not allow any direct references to it to remain accessible.
	 */
	public static <B> MutableSubClassToInstanceMap<B>
			create(Map<Class<? extends B>, B> backingMap) {
		return new MutableSubClassToInstanceMap<B>(backingMap);
	}

	/** Constructor. */
	private MutableSubClassToInstanceMap(Map<Class<? extends B>, B> delegate) {
		this.delegate = delegate;
	}

	/** {@inheritDoc} */
	@Override
	public <T extends B> T putInstance(Class<T> type, T value) {
		return cast(type, put(type, value));
	}

	/** {@inheritDoc} */
	@SuppressWarnings({"unchecked", "rawtypes"})
	@Override
	public <T extends B> T getInstance(Class<T> type) {
		T retVal = cast(type, get(type));
		if(retVal == null) {
			// The raw map cast is extremely ugly, but we ensure beforehand that the delegate map
			// contains only types which we want.
			return (T)performNearestClassLookup(type, (Map)delegate);
		}
		return retVal;
	}

	/** Support method for using primitive types in this map. */
	private static <B, T extends B> T cast(Class<T> type, B value) {
		return Primitives.wrap(type).cast(value);
	}

	/** {@inheritDoc} */
	@Override
	protected Map<Class<? extends B>, B> delegate() {
		return delegate;
	}
}
