/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.module;

import static org.fortiss.af3.exploration.util.DesignSpaceExplorationModelElementFactory.createSuperSetMap;

import java.util.HashSet;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.dsl_v2.model.expression.SuperSet;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.tooling.base.model.element.IModelElement;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.MutableClassToInstanceMap;

/**
 * Container class that provides inputs from a DSE to the transformation framework.
 * 
 * @author diewald
 */
// TODO: Instead of wrapping, we could also use subclassing: Encodings for DSE internal instances,
// SuperSets for the external world.
public class ExplorationTransformationInputs {
	/** References the input {@link SuperSet}s. */
	private SuperSetMap superSets;
	/** References the internal inputs from the DSE. */
	// TODO(AD): This explicit declaration of the Phenotype parameter can be avoided by
	// subclassing...
	private ClassToInstanceMap<Phenotype> dseEncodings;

	/** Constructor. */
	public ExplorationTransformationInputs(SuperSetMap superSets,
			ClassToInstanceMap<Phenotype> dseEncodings) {
		if(superSets != null) {
			this.superSets = superSets;
		} else {
			this.superSets = createSuperSetMap();
		}
		if(dseEncodings != null) {
			this.dseEncodings = dseEncodings;
		} else {
			this.dseEncodings = MutableClassToInstanceMap.create();
		}
	}

	/**
	 * Returns an instance of the given class from the encodings of a DSE.
	 * 
	 * @param type
	 *            of the instance to be returned.
	 * @return instance of the class obtained from the above mentioned containers, or {@code null}
	 *         if it is not found.
	 */
	public <T extends Phenotype> T getEncoding(Class<T> type) {
		return dseEncodings.getInstance(type);
	}

	/**
	 * Returns the {@link SuperSet} which references elements of the given type.
	 * 
	 * @param type
	 *            of the {@link SuperSet} to be returned.
	 * @return {@link SuperSet} of the given type or {@code null} if no corresponding
	 *         {@link SuperSet} is not found.
	 */
	public <S extends IModelElement> SuperSet<S> getSuperSet(Class<S> type) {
		return superSets.get(type);
	}

	/**
	 * Returns the types of all instances of DSE results wrapped in this container (encodings or
	 * {@link SuperSet}s).
	 */
	public Set<Class<?>> keySet() {
		Set<Class<?>> allTypes = new HashSet<>();
		allTypes.addAll(dseEncodings.keySet());
		allTypes.addAll(superSets.keySet());
		return allTypes;
	}
}
