/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSuperSet;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.List;

import org.fortiss.af3.exploration.dsl_v2.model.booleanp.Exists;
import org.fortiss.af3.exploration.dsl_v2.model.booleanp.ForAll;
import org.fortiss.af3.exploration.dsl_v2.model.booleanp.allocation.Allocation;
import org.fortiss.af3.exploration.dsl_v2.model.booleanp.allocation.ILocationExpression;
import org.fortiss.af3.exploration.dsl_v2.model.expression.IExpression;
import org.fortiss.af3.exploration.dsl_v2.model.expression.SuperSet;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.task.model.Task;

/**
 * Utility methods for handling DSML {@link IExpression}s of {@link ExplorationTarget}s
 * 
 * @author diewald
 */
public class DSMLUtils {

	/**
	 * Extracts the {@link ILocationExpression} of an allocation pattern
	 * {@link ExplorationConstraint} or {@code null} if the {@link ExplorationConstraint} is not
	 * such a pattern.
	 */
	public static ILocationExpression
			getDirectAllocationExpression(ExplorationConstraint<?> constraint) {
		IExpression expr = constraint.getExpression();
		if(!(expr instanceof ForAll)) {
			return null;
		}
		IExpression forAllExpr = ((ForAll)expr).getExpression();
		if(!(forAllExpr instanceof Exists)) {
			return null;
		}
		IExpression existsExpr = ((Exists)forAllExpr).getExpression();
		if(existsExpr instanceof ILocationExpression) {
			return (ILocationExpression)existsExpr;
		}
		return null;
	}

	/** Returns whether the given {@link ExplorationConstraint} is an allocation pattern. */
	public static boolean isAllocationPattern(ExplorationConstraint<?> constraint) {
		return getDirectAllocationExpression(constraint) != null;
	}

	/**
	 * From a given ExplorationSpecification, this method examines all
	 * {@link ExplorationConstraint}s for an allocation pattern where all tasks from the input
	 * {@link SuperSet} are referenced. This constraint is interpreted as a limitation of the
	 * "allowed" mapping targets in hierarchical platforms. The referenced allocation targets are
	 * returned as a new {@link SuperSet} that can be forwarded to a solver as the target
	 * definition.
	 * 
	 * @param allTasks
	 *            {@link SuperSet} of the tasks to be mapped.
	 * @param expSpec
	 *            input specification to the solver.
	 * @return {@link SuperSet} of mapping targets.
	 */
	public static SuperSet<ExecutionUnit> extractTargetSuperSetFromAllocConstr(
			SuperSet<Task> allTasks, ExplorationSpecification expSpec) {
		// Avoid the detection of allocation operators in the objectives...
		@SuppressWarnings("unchecked") List<ExplorationConstraint<?>> constraints =
				(List<ExplorationConstraint<?>>)(List<?>)pickInstanceOf(ExplorationConstraint.class,
						expSpec.getTargets());
		for(ExplorationConstraint<?> constr : constraints) {
			ILocationExpression locExpression = getDirectAllocationExpression(constr);
			if(locExpression instanceof Allocation) {
				// See the conditional above, safe access with get(0).
				Allocation candidateAlloc = (Allocation)locExpression;
				org.fortiss.af3.exploration.dsl_v2.model.expression.Set<?> leftSet =
						candidateAlloc.getLeft().getSetReference();
				if(leftSet.getEntryType() == Task.class) {
					if(allTasks.getEntries().size() == leftSet.getEntries().size() &&
							leftSet.getEntries().containsAll(allTasks.getEntries())) {
						org.fortiss.af3.exploration.dsl_v2.model.expression.Set<?> rightSet =
								candidateAlloc.getRight().getSetReference();
						if(rightSet.getEntryType() == ExecutionUnit.class) {
							return createSuperSet(
									rightSet.getCastedSet(ExecutionUnit.class).getEntries(),
									"TargetExecUnits", ExecutionUnit.class);
						}
					}
				}
			}
		}
		return null;
	}
}
