/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.component.utils.ComponentModelElementFactory.createChannelAndAttach;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.af3.allocation.model.AllocationEntry;
import org.fortiss.af3.allocation.model.ManyToOneAllocationEntry;
import org.fortiss.af3.allocation.model.OneToManyAllocationEntry;
import org.fortiss.af3.allocation.model.OneToOneAllocationEntry;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.utils.ComponentModelElementFactory;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.dsl_v2.model.expression.SuperSet;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.base.utils.OffsetOrientationUtils;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;

import com.google.common.collect.Multimap;

/**
 * Utility methods for transforming DSE-internal models to AF3 models.
 * 
 * @author diewald
 */
public class TransformationUtils {

	/** Constructor. Prevent initialization. */
	private TransformationUtils() {
	}

	/**
	 * Returns the architecture of the given type that contains one of the elements referenced by an
	 * {@link AllocationEntry} of the given {@link SuperSet}.
	 * 
	 * @param superSet
	 *            containing allocations with references to elements of an architecture of the given
	 *            type.
	 * @param architectureType
	 *            type of the containing architecture.
	 * @return The containing architecture, or {@code null} if the set references only elements that
	 *         are contained in no {@link IProjectRootElement} or an architecture of a different
	 *         type.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends IProjectRootElement> T getArchitectureOfAllocSuperSet(
			SuperSet<? extends AllocationEntry> superSet, Class<T> architectureType) {
		AllocationEntry entry = getFirst(superSet.getEntries()).orElse(null);
		if(entry == null) {
			return null;
		}

		IProjectRootElement architectureModel = null;

		if(entry instanceof OneToOneAllocationEntry) {
			IModelElement element = ((OneToOneAllocationEntry)entry).getSourceElement();
			architectureModel = getFirstParentWithType(element, architectureType);
			if(architectureModel != null) {
				return (T)architectureModel;
			}
			element = ((OneToOneAllocationEntry)entry).getTargetElement();
			architectureModel = getFirstParentWithType(element, architectureType);
			return (T)architectureModel;
		} else if(entry instanceof OneToManyAllocationEntry) {
			IModelElement element = ((OneToManyAllocationEntry)entry).getSourceElement();
			architectureModel = getFirstParentWithType(element, architectureType);
			if(architectureModel != null) {
				return (T)architectureModel;
			}
			element = getFirst(((OneToManyAllocationEntry)entry).getTargetElements()).orElse(null);
			architectureModel = getFirstParentWithType(element, architectureType);
			return (T)architectureModel;
		} else if(entry instanceof ManyToOneAllocationEntry) {
			IModelElement element =
					getFirst(((ManyToOneAllocationEntry)entry).getSourceElements()).orElse(null);
			architectureModel = getFirstParentWithType(element, architectureType);
			if(architectureModel != null) {
				return (T)architectureModel;
			}
			element = ((ManyToOneAllocationEntry)entry).getTargetElement();
			architectureModel = getFirstParentWithType(element, architectureType);
			return (T)architectureModel;
		}
		return null;
	}

	/**
	 * Returns a list of target ports within the given {@link ComponentArchitecture}. Target ports
	 * are defined as ports that do not have direct outputs.
	 * 
	 * @param componentArchitecture
	 *            The {@link ComponentArchitecture} for whom source ports shall be identified.
	 * @return {@link EList} of target {@link InputPort}s.
	 */
	public static EList<InputPort> getTargetPorts(ComponentArchitecture componentArchitecture) {
		if(componentArchitecture == null) {
			return null;
		}

		EList<InputPort> allOutputPortsInArch =
				getChildrenWithType(componentArchitecture, InputPort.class);
		EList<InputPort> targetPortList = new BasicEList<InputPort>();
		for(InputPort currentOutputPort : allOutputPortsInArch) {
			// Filter out ports that represent the transition through a hierarchic hull.
			if(currentOutputPort.getOutgoing().isEmpty()) {
				targetPortList.add(currentOutputPort);
			}
		}

		return targetPortList;
	}

	/**
	 * Creates an identical copy of a given {@link InputPort}. Used for the replication of logical
	 * architectures.
	 * 
	 * @param referencePort
	 *            The {@link InputPort} to be replicated.
	 * @param replicationOwner
	 *            The {@link Component} to which the replicated {@link InputPort} shall be attached
	 *            to.
	 * @param portName
	 *            The name of the replicated port {@link InputPort} (NB: It must be different to the
	 *            original name if attached to the same {@link Component}. Otherwise the containing
	 *            architecture is invalid.
	 * @return The replicated {@link InputPort}.
	 */
	public static InputPort replicateInputPortAndAttach(Port referencePort,
			Component replicationOwner, String portName) {
		InputPort replicatedInputPort =
				ComponentModelElementFactory.createInputPort(portName, referencePort.getComment());
		replicatedInputPort.getSpecifications().clear();
		replicatedInputPort.getSpecifications()
				.addAll(EcoreUtil.copyAll(referencePort.getSpecifications()));

		replicationOwner.getConnectors().add(replicatedInputPort);

		return replicatedInputPort;
	}

	/**
	 * Creates an identical copy of a given {@link OutputPort}. Used for the replication of logical
	 * architectures.
	 * 
	 * @param referencePort
	 *            The {@link OutputPort} to be replicated.
	 * @param replicationOwner
	 *            The {@link Component} to which the replicated {@link OutputPort} shall be attached
	 *            to.
	 * @param portName
	 *            The name of the replicated port {@link OutputPort} (NB: It must be different to
	 *            the original name if attached to the same {@link Component}. Otherwise the
	 *            containing architecture is invalid.
	 * @return The replicated {@link OutputPort}.
	 */
	public static OutputPort replicateOutputPortAndAttach(Port referencePort,
			Component replicationOwner, String portName) {
		OutputPort replicatedOutputPort =
				ComponentModelElementFactory.createOutputPort(portName, referencePort.getComment());
		replicatedOutputPort.getSpecifications().clear();
		replicatedOutputPort.getSpecifications()
				.addAll(EcoreUtil.copyAll(referencePort.getSpecifications()));

		replicationOwner.getConnectors().add(replicatedOutputPort);

		return replicatedOutputPort;
	}

	/**
	 * Sets the position of the generated {@link Component} with some offset to the position of the
	 * original, i.e. source, {@link Component}.
	 */
	public static void setGenComponentPos(Component sourceComponent, Component generatedComponent) {
		// get(0): Each component has a layout data if they were constructed as intended.
		if(generatedComponent != null &&
				generatedComponent.getLayoutData().get(0) instanceof Point &&
				sourceComponent.getLayoutData().get(0) instanceof Point) {
			// get(0): Each layout data of a component has a point if they were constructed as
			// intended.
			Point p = (Point)sourceComponent.getLayoutData().get(0);
			p = calulateOffset(p, sourceComponent, 5, false, false);
			generatedComponent.getLayoutData().set(0, p);
		}
	}

	/**
	 * Adds the suffix that denotes the replication number to the given instantiated element.
	 * 
	 * @param origElement
	 *            Element of the original input model.
	 * @param abstractElement
	 *            "Abstract", replicated, element from the generated model.
	 * @param instantiatedElement
	 *            Instantiated element from the element library.
	 */
	public static void setReplicaAwareInstElementName(INamedElement origElement,
			INamedElement abstractElement, INamedElement instantiatedElement) {
		String suffix = abstractElement.getName().replace(origElement.getName(), "");
		instantiatedElement.setName(instantiatedElement.getName() + suffix);
	}

	/**
	 * Calculates a point (at which a component can be placed) by the using the information about
	 * the {@link Dimension}s of the {#code refenceComponent} and an additional offset. The boolean
	 * parameters are used to determine whether the dimension of the {@code referenceComponent} in
	 * the one or the other direction shall be taken into account.
	 * NOTE: negative values in the parameter {#code offsetDistance} cause a shift to the upper
	 * left, otherwise to the lower right.
	 */
	public static Point calulateOffset(Point pointToMove, Component referenceComponent,
			int offsetDistance, boolean horizontalOffset, boolean verticalOffset) {
		int xOffsetToReferenceComponent = 0;
		int yOffsetToReferenceComponent = 0;

		// Retrieve the information about the dimensions of the parent Component, if this
		// information is relevant for the shift.
		ILayoutData dimensionOfReplicatedComponent = referenceComponent.getLayoutData().get(1);
		if(dimensionOfReplicatedComponent instanceof Dimension) {
			if(horizontalOffset) {
				xOffsetToReferenceComponent =
						((Dimension)dimensionOfReplicatedComponent).getWidth();
			}
			if(verticalOffset) {
				yOffsetToReferenceComponent =
						((Dimension)dimensionOfReplicatedComponent).getHeight();
			}
		}

		// Consider the dimensions of the referenced Component as an additional shift to the left,
		// if the offset variable defines a shift to the left.
		if(offsetDistance < 0) {
			xOffsetToReferenceComponent = -xOffsetToReferenceComponent;
			yOffsetToReferenceComponent = -yOffsetToReferenceComponent;
		}

		// Calculate the coordinates of the shifted point that will be returned.
		pointToMove.setX(pointToMove.getX() + xOffsetToReferenceComponent + offsetDistance);
		pointToMove.setY(pointToMove.getY() + yOffsetToReferenceComponent + offsetDistance);

		return pointToMove;
	}

	/**
	 * Creates a set of {@link Channel}s that connects the given {#link generatedPort} with the
	 * {#link connectedPorts}. The given {#link parentComponent} will contain the resulting set of
	 * {@link Channel}s.
	 */
	public static <T extends Port, S extends Port> void connectPort(T generatedPort,
			Collection<S> connectedPorts, Component parentComponent) {
		for(S currentPortToBeConnected : connectedPorts) {
			// Depending on the port type (input, output), we need to swap the creation of the
			// Channel's sources and targets.
			if(generatedPort instanceof InputPort) {
				createChannelAndAttach(parentComponent.getParentComponent(), "",
						currentPortToBeConnected, generatedPort);
			} else if(generatedPort instanceof OutputPort) {
				createChannelAndAttach(parentComponent.getParentComponent(), "", generatedPort,
						currentPortToBeConnected);
			}
		}
	}

	/**
	 * Returns the set of {@link ITaskAdapter}s that is defined by the given
	 * {@link IMappingEncoding}. This method is intended to be used within
	 * {@link ITransformationModule}s.
	 * 
	 * @param tModule
	 *            Calling {@link ITransformationModule}.
	 * @param mapping
	 *            {@link IMappingEncoding} that encodes the deployed
	 *            {@link ITaskAdapter}s.
	 * @return Collection of deployable {@link ITaskAdapter}s.
	 * @throws TransformationModuleException
	 *             If one {@link ITaskAdapter} of the mapping does not wrap an AF3
	 *             {@link Component} but any other object.
	 */
	public static Collection<ITaskAdapter<Component>> retrieveDeployableComponents(
			ITransformationModule<?> tModule, AbstractTaskMappingEncoding mapping)
			throws TransformationModuleException {
		Collection<ITaskAdapter<Component>> deplCompAdps = new ArrayList<>();
		for(ITaskAdapter<?> comp : mapping.getRequesters()) {
			if(comp.getObject() instanceof Component) {
				@SuppressWarnings("unchecked") ITaskAdapter<Component> castComp =
						(ITaskAdapter<Component>)comp;
				deplCompAdps.add(castComp);
			} else {
				throw new TransformationModuleException(tModule,
						"Error. The ComponentAdapter from the DSE is not an adapter for an AF3 component.");
			}
		}

		return deplCompAdps;
	}

	/**
	 * Returns a {@link Map} that associates the Components of the given Architecture with the
	 * {@link ITaskAdapter}s from the exploration.
	 * 
	 * @throws TransformationModuleException
	 *             if a {@link Component} from the original {@link ComponentArchitecture} cannot be
	 *             associated with one, but only one, Component in the output architecture.
	 */
	// TODO: update doc.
	public static Map<ITaskAdapter<Component>, Component> getDeployableComponentMap(
			ITransformationModule<?> tModule, Multimap<EObject, EObject> origToGenAssocMap,
			Collection<ITaskAdapter<Component>> deplCompCollection)
			throws TransformationModuleException {
		Map<ITaskAdapter<Component>, Component> deployableComponentsOfReplicatedArch =
				new ConcurrentHashMap<>();
		for(ITaskAdapter<Component> deployableComponent : deplCompCollection) {
			Component correspondingAF3Component = deployableComponent.getObject();
			if(deployableComponent.isInstantiation()) {
				correspondingAF3Component = deployableComponent.getReplacedComponent().getObject();
			}
			Collection<EObject> genComponentColl = origToGenAssocMap.get(correspondingAF3Component);
			if(genComponentColl == null || genComponentColl.isEmpty()) {
				throw new TransformationModuleException(tModule,
						"No generated Component that is association with the Component " +
								correspondingAF3Component.getName() + " has been found.");
			} else if(genComponentColl.size() > 1) {
				throw new TransformationModuleException(tModule,
						"More than one generated Component that are associated with the original Component " +
								correspondingAF3Component.getName() +
								" have been found. This is not supported in this stage of the transformation.");
			}

			Component deployableAF3Component = (Component)genComponentColl.stream().findAny().get();
			deployableComponentsOfReplicatedArch.put(deployableComponent, deployableAF3Component);
		}
		return deployableComponentsOfReplicatedArch;
	}

	/**
	 * Arranges the given {@link Port}s such that they are placed with equal distances either on
	 * left or the right side of the given {@link Component}. The given {@link Component} must be
	 * the owner of all ports given in the list.
	 */
	public static void arrangePorts(EList<? extends Port> connectorsToArrange,
			ILayoutedModelElement parentElement) {
		if(parentElement.getLayoutData().get(1) instanceof Dimension) {
			int voterHeight = ((Dimension)parentElement.getLayoutData().get(1)).getHeight();
			float distanceBetweenPorts = voterHeight / (connectorsToArrange.size() + 1);
			float currentPortOffset = distanceBetweenPorts / 2;

			for(Port currentPort : connectorsToArrange) {
				EOrientation inOutOrientation;
				if(currentPort instanceof OutputPort) {
					inOutOrientation = EOrientation.EAST;
				} else {
					inOutOrientation = EOrientation.WEST;
				}
				OffsetOrientationUtils.setOffsetOrientation(currentPort, "coffsetorientation",
						(int)currentPortOffset, inOutOrientation);

				currentPortOffset = currentPortOffset + distanceBetweenPorts;
			}
		}
	}
}
